"""
Model wrapper.
"""

# This file is part of Echofilter.
#
# Copyright (C) 2020-2022  Scott C. Lowe and Offshore Energy Research Association (OERA)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import warnings

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.modules.loss import _Loss

from .utils import TensorDict, logavgexp


class Echofilter(nn.Module):
    """
    Echofilter logit mapping wrapper.

    Parameters
    ----------
    model : :class:`torch.nn.Module`
        The model backbone, which converts inputs to logits.
    top : str, optional
        Type of output for top line and surface line. If ``"mask"``, the top
        output corresponds to logits, which are converted into probabilities
        with sigmoid. If ``"boundary"`` (default), the output corresponds to
        logits for the location of the line, which is converted into a
        probability mask using softmax and cumsum.
    bottom : str, optional
        As for ``top``, but for the bottom line. Default is ``"boundary"``.
    mapping : dict or None, optional
        Mapping from logit names to output channels provided by ``model``.
        If ``None``, a default mapping is used. The mapping is stored as
        ``self.mapping``.
    reduction_ispassive : str, default="logavgexp"
        Method used to reduce the depths dimension for the ``"logit_is_passive"``
        output.
    reduction_isremoved : str , default="logavgexp"
        Method used to reduce the depths dimension for the ``"logit_is_removed"``
        output.
    conditional : bool, optional
        Whether to build a conditional model as well as an unconditional model.
        If ``True``, there are additional logits in the call output named
        ``"x|downfacing"`` and ``"x|upfacing"``, in addition to
        ``"x"``. For instance, ``"p_is_above_turbulence|downfacing"``. Default is ``False``.
    """

    aliases = [("top", "turbulence")]

    def __init__(
        self,
        model,
        top="boundary",
        bottom="boundary",
        mapping=None,
        reduction_ispassive="logavgexp",
        reduction_isremoved="logavgexp",
        conditional=False,
    ):
        super(Echofilter, self).__init__()
        self.model = model
        self.params = {
            "top": top,
            "bottom": bottom,
            "reduction_ispassive": reduction_ispassive,
            "reduction_isremoved": reduction_isremoved,
            "conditional": conditional,
        }
        if mapping is None:
            mapping = {
                "logit_is_above_turbulence": 0,
                "logit_is_boundary_turbulence": 0,
                "logit_is_below_bottom": 1,
                "logit_is_boundary_bottom": 1,
                "logit_is_removed": 2,
                "logit_is_passive": 3,
                "logit_is_patch": 4,
                "logit_is_above_surface": 5,
                "logit_is_boundary_surface": 5,
                "logit_is_above_turbulence-original": 6,
                "logit_is_boundary_turbulence-original": 6,
                "logit_is_below_bottom-original": 7,
                "logit_is_boundary_bottom-original": 7,
                "logit_is_patch-original": 8,
                "logit_is_patch-ntob": 9,
            }
            if top == "boundary":
                mapping.pop("logit_is_above_turbulence")
                mapping.pop("logit_is_above_turbulence-original")
                mapping.pop("logit_is_above_surface")
            else:
                mapping.pop("logit_is_boundary_turbulence")
                mapping.pop("logit_is_boundary_turbulence-original")
                mapping.pop("logit_is_boundary_surface")
            if bottom == "boundary":
                mapping.pop("logit_is_below_bottom")
                mapping.pop("logit_is_below_bottom-original")
            else:
                mapping.pop("logit_is_boundary_bottom")
                mapping.pop("logit_is_boundary_bottom-original")
        self.mapping = mapping
        # Ensure all references for aliases are set
        mapping_extra = {}
        for key in mapping:
            for alias_map in self.aliases:
                for (alias_a, alias_b) in [alias_map, alias_map[::-1]]:
                    if "_" + alias_a not in key:
                        continue
                    alt_key = key.replace("_" + alias_a, "_" + alias_b)
                    if alt_key not in mapping:
                        mapping_extra[alt_key] = mapping[key]
        mapping.update(mapping_extra)

        self.conditions = [""]
        if conditional:
            self.conditions += ["downfacing", "upfacing"]
        self.n_outputs_per_condition = max(self.mapping.values())

    def forward(self, x, output_device=None):
        logits = self.model(x)
        outputs = TensorDict()

        for i_condition, condition in enumerate(self.conditions):
            # Define the condition string
            cs = condition
            if cs != "":
                cs = "|" + cs
            # Define the logit index offset
            ofs = i_condition * self.n_outputs_per_condition

            # Include raw logits in output
            for key, index in self.mapping.items():
                outputs[key + cs] = logits[:, index + ofs]

            # Flatten some outputs which are vectors not arrays
            if self.params["reduction_isremoved"] == "mean":
                outputs["logit_is_removed" + cs] = torch.mean(
                    outputs["logit_is_removed" + cs], dim=-1
                ).to(device=output_device)
            elif self.params["reduction_isremoved"] in {"logavgexp", "lae"}:
                outputs["logit_is_removed" + cs] = logavgexp(
                    outputs["logit_is_removed" + cs], dim=-1
                ).to(device=output_device)
            else:
                raise ValueError(
                    "Unsupported reduction_isremoved value: {}".format(
                        self.params["reduction_isremoved"]
                    )
                )

            if self.params["reduction_ispassive"] == "mean":
                outputs["logit_is_passive" + cs] = torch.mean(
                    outputs["logit_is_passive" + cs], dim=-1
                ).to(device=output_device)
            elif self.params["reduction_ispassive"] in {"logavgexp", "lae"}:
                outputs["logit_is_passive" + cs] = logavgexp(
                    outputs["logit_is_passive" + cs], dim=-1
                ).to(device=output_device)
            else:
                raise ValueError(
                    "Unsupported reduction_ispassive value: {}".format(
                        self.params["reduction_ispassive"]
                    )
                )

            # Convert logits to probabilities
            outputs["p_is_removed" + cs] = torch.sigmoid(
                outputs["logit_is_removed" + cs]
            ).to(device=output_device)
            outputs["p_is_passive" + cs] = torch.sigmoid(
                outputs["logit_is_passive" + cs]
            ).to(device=output_device)

            for sfx in ("turbulence", "turbulence-original", "surface"):
                if self.params["top"] == "mask":
                    outputs["p_is_above_" + sfx + cs] = torch.sigmoid(
                        outputs["logit_is_above_" + sfx + cs]
                    ).to(device=output_device)
                    outputs["p_is_below_" + sfx + cs] = (
                        1 - outputs["p_is_above_" + sfx + cs]
                    ).to(device=output_device)
                elif self.params["top"] == "boundary":
                    outputs["p_is_boundary_" + sfx + cs] = F.softmax(
                        outputs["logit_is_boundary_" + sfx + cs], dim=-1
                    ).to(device=output_device)
                    outputs["p_is_above_" + sfx + cs] = torch.flip(
                        torch.cumsum(
                            torch.flip(
                                outputs["p_is_boundary_" + sfx + cs], dims=(-1,)
                            ),
                            dim=-1,
                        ),
                        dims=(-1,),
                    ).to(device=output_device)
                    outputs["p_is_below_" + sfx + cs] = torch.cumsum(
                        outputs["p_is_boundary_" + sfx + cs], dim=-1
                    ).to(device=output_device)
                    # Due to floating point precision, max value can exceed 1.
                    # Fix this by clipping the values to the appropriate range.
                    outputs["p_is_above_" + sfx + cs].clamp_(0, 1)
                    outputs["p_is_below_" + sfx + cs].clamp_(0, 1)
                else:
                    raise ValueError(
                        'Unsupported "top" parameter: {}'.format(self.params["top"])
                    )

            for sfx in ("bottom", "bottom-original"):
                if self.params["bottom"] == "mask":
                    outputs["p_is_below_" + sfx + cs] = torch.sigmoid(
                        outputs["logit_is_below_" + sfx + cs]
                    ).to(device=output_device)
                    outputs["p_is_above_" + sfx + cs] = (
                        1 - outputs["p_is_below_" + sfx + cs]
                    ).to(device=output_device)
                elif self.params["bottom"] == "boundary":
                    outputs["p_is_boundary_" + sfx + cs] = F.softmax(
                        outputs["logit_is_boundary_" + sfx + cs], dim=-1
                    ).to(device=output_device)
                    outputs["p_is_below_" + sfx + cs] = torch.cumsum(
                        outputs["p_is_boundary_" + sfx + cs], dim=-1
                    ).to(device=output_device)
                    outputs["p_is_above_" + sfx + cs] = torch.flip(
                        torch.cumsum(
                            torch.flip(
                                outputs["p_is_boundary_" + sfx + cs], dims=(-1,)
                            ),
                            dim=-1,
                        ),
                        dims=(-1,),
                    ).to(device=output_device)
                    # Due to floating point precision, max value can exceed 1.
                    # Fix this by clipping the values to the appropriate range.
                    outputs["p_is_below_" + sfx + cs].clamp_(0, 1)
                    outputs["p_is_above_" + sfx + cs].clamp_(0, 1)
                else:
                    raise ValueError(
                        'Unsupported "bottom" parameter: {}'.format(
                            self.params["bottom"]
                        )
                    )

            for sfx in ("", "-original", "-ntob"):
                outputs["p_is_patch" + sfx + cs] = torch.sigmoid(
                    outputs["logit_is_patch" + sfx + cs]
                ).to(device=output_device)

            outputs["p_keep_pixel" + cs] = (
                1.0
                * 0.5
                * (
                    (1 - outputs["p_is_above_turbulence" + cs])
                    + outputs["p_is_below_turbulence" + cs]
                )
                * 0.5
                * (
                    (1 - outputs["p_is_below_bottom" + cs])
                    + outputs["p_is_above_bottom" + cs]
                )
                * (1 - outputs["p_is_removed" + cs].unsqueeze(-1))
                * (1 - outputs["p_is_passive" + cs].unsqueeze(-1))
                * (1 - outputs["p_is_patch" + cs])
            ).clamp_(0, 1)
            outputs["mask_keep_pixel" + cs] = (
                1.0
                * (outputs["p_is_above_turbulence" + cs] < 0.5)
                * (outputs["p_is_below_bottom" + cs] < 0.5)
                * (outputs["p_is_removed" + cs].unsqueeze(-1) < 0.5)
                * (outputs["p_is_passive" + cs].unsqueeze(-1) < 0.5)
                * (outputs["p_is_patch" + cs] < 0.5)
            )
        return outputs


class EchofilterLoss(_Loss):
    """
    Evaluate loss for an Echofilter model.

    Parameters
    ----------
    reduction : ``"mean"`` or ``"sum"``, optional
        The reduction method, which is used to collapse batch and timestamp
        dimensions. Default is ``"mean"``.
    turbulence_mask : float, optional
        Weighting for turbulence line/mask loss term. Default is ``1.0``.
    bottom_mask : float, optional
        Weighting for bottom line/mask loss term. Default is ``1.0``.
    removed_segment : float, optional
        Weighting for ``is_removed`` loss term. Default is ``1.0``.
    passive : float, optional
        Weighting for ``is_passive`` loss term. Default is ``1.0``.
    patch : float, optional
        Weighting for ``mask_patch`` loss term. Default is ``1.0``.
    overall : float, optional
        Weighting for overall mask loss term. Default is ``0.0``.
    surface : float, optional
        Weighting for surface line/mask loss term. Default is ``1.0``.
    auxiliary : float, optional
        Weighting for auxiliary loss terms ``"turbulence-original"``,
        ``"bottom-original"``, ``"mask_patches-original"``, and
        ``"mask_patches-ntob"``. Default is ``1.0``.
    ignore_lines_during_passive : bool, optional
        Whether targets for turbulence and bottom lines should be excluded from
        the loss during passive data collection. Default is ``True``.
    ignore_lines_during_removed : bool, optional
        Whether targets for turbulence and bottom lines should be excluded from
        the loss during entirely removed sections. Default is ``True``.
    ignore_surface_during_passive : bool, optional
        Whether target for the surface line should be excluded from the loss
        during passive data collection. Default is ``False``.
    ignore_surface_during_removed : bool, optional
        Whether target for the surface line should be excluded from the loss
        during entirely removed sections. Default is ``True``.
    """

    __constants__ = ["reduction"]

    def __init__(
        self,
        reduction="mean",
        conditional=False,
        turbulence_mask=1.0,
        bottom_mask=1.0,
        removed_segment=1.0,
        passive=1.0,
        patch=1.0,
        overall=0.0,
        surface=1.0,
        auxiliary=1.0,
        ignore_lines_during_passive=False,
        ignore_lines_during_removed=True,
        ignore_surface_during_passive=False,
        ignore_surface_during_removed=True,
    ):
        super(EchofilterLoss, self).__init__(None, None, reduction)
        self.conditional = conditional
        self.turbulence_mask = turbulence_mask
        self.bottom_mask = bottom_mask
        self.removed_segment = removed_segment
        self.passive = passive
        self.patch = patch
        self.overall = overall
        self.surface = surface
        self.auxiliary = auxiliary
        self.ignore_lines_during_passive = ignore_lines_during_passive
        self.ignore_lines_during_removed = ignore_lines_during_removed
        self.ignore_surface_during_passive = ignore_surface_during_passive
        self.ignore_surface_during_removed = ignore_surface_during_removed

        self.conditions = [""]
        if conditional:
            self.conditions += ["downfacing", "upfacing"]

    def forward(self, input, target):
        """
        Construct loss term.

        Parameters
        ----------
        input : dict
            Output from :class:`echofilter.wrapper.Echofilter` layer.
        target : dict
            A transect, as provided by :class:`echofilter.data.dataset.TransectDataset`.
        """
        loss = 0

        target["is_passive"] = target["is_passive"].to(
            input["logit_is_passive"].device,
            input["logit_is_passive"].dtype,
        )
        target["is_removed"] = target["is_removed"].to(
            input["logit_is_removed"].device,
            input["logit_is_removed"].dtype,
        )

        n_conditions_in_loss = 0
        for condition in self.conditions:
            closs = 0
            with torch.no_grad():
                if condition == "":
                    cs = condition
                    cmask = torch.ones_like(target["is_upward_facing"])
                else:
                    cs = "|" + condition
                    if condition == "upfacing":
                        cmask = target["is_upward_facing"] > 0.5
                    elif condition == "downfacing":
                        cmask = target["is_upward_facing"] < 0.5
                    else:
                        raise ValueError("Unsupported condition: {}".format(condition))
                    n_samples_in_condition = torch.sum(cmask)
                    if n_samples_in_condition.cpu().item() == 0:
                        # No samples in this batch match this condition
                        continue
                cmask = cmask.to(torch.float32)

            n_conditions_in_loss += 1

            for sfx in ("turbulence", "turbulence-original", "surface"):
                with torch.no_grad():
                    loss_inclusion_mask = (target["is_bad_labels"] < 1e-7).to(
                        torch.float32
                    )
                    if sfx == "surface":
                        target_key = "mask_surface"
                        target_i_key = "index_surface"
                        weight = self.surface
                        # Don't include surrogate surface datapoints in loss
                        loss_inclusion_mask *= (
                            target["is_surrogate_surface"] < 1e-7
                        ).to(torch.float32)
                        # Check whether surface line is masked out
                        if self.ignore_surface_during_passive:
                            loss_inclusion_mask *= 1 - target["is_passive"]
                        if self.ignore_surface_during_removed:
                            loss_inclusion_mask *= 1 - target["is_removed"]
                    else:
                        target_key = "mask_" + sfx
                        target_i_key = "index_" + sfx
                        weight = self.turbulence_mask
                        if sfx != "turbulence":
                            weight *= self.auxiliary
                        # Check whether line is masked out
                        if self.ignore_lines_during_passive:
                            loss_inclusion_mask *= 1 - target["is_passive"]
                        if self.ignore_lines_during_removed:
                            loss_inclusion_mask *= 1 - target["is_removed"]

                if not weight:
                    continue
                elif "logit_is_boundary_" + sfx in input:
                    # Load cross-entropy class target
                    C = target[target_i_key].to(
                        device=input["logit_is_boundary_" + sfx + cs].device,
                        dtype=torch.long,
                    )
                    loss_term = F.cross_entropy(
                        input["logit_is_boundary_" + sfx + cs].transpose(-2, -1),
                        C,
                        reduction="none",
                    )
                    loss_term *= cmask.unsqueeze(-1)
                    loss_term *= loss_inclusion_mask
                    if self.reduction == "mean":
                        loss_term = torch.mean(loss_term)
                    elif self.reduction == "sum":
                        loss_term = torch.sum(loss_term)
                    elif self.reduction != "none":
                        raise ValueError(
                            "Unsupported reduction: {}".format(self.reduction)
                        )
                elif "logit_is_above_" + sfx in input:
                    warnings.warn(
                        'Using loss corresponding to "mask" logits.'
                        ' The "boundary" is recommended instead.'
                        " The loss component for this line will be"
                        " F.binary_cross_entropy_with_logits(input[{}], target[{}])"
                        "".format("logit_is_above_" + sfx + cs, target_key)
                    )
                    loss_term = F.binary_cross_entropy_with_logits(
                        input["logit_is_above_" + sfx + cs],
                        target[target_key].to(
                            input["logit_is_above_" + sfx + cs].device,
                            input["logit_is_above_" + sfx + cs].dtype,
                        ),
                        reduction="none",
                    )
                    loss_term *= cmask.unsqueeze(-1).unsqueeze(-1)
                    loss_term *= loss_inclusion_mask.unsqueeze(-1)
                    if self.reduction == "mean":
                        loss_term = torch.mean(loss_term)
                    elif self.reduction == "sum":
                        loss_term = torch.sum(loss_term)
                    elif self.reduction != "none":
                        raise ValueError(
                            "Unsupported reduction: {}".format(self.reduction)
                        )
                else:
                    raise ValueError(
                        "The input does not contain either {} or {} fields."
                        " At least one of these is required if the loss term weighting"
                        " is non-zero.".format(
                            "logit_is_boundary_" + sfx,
                            "logit_is_above_" + sfx,
                        )
                    )
                if torch.isnan(loss_term).any():
                    print("Loss term {} is NaN".format(target_key))
                else:
                    closs += weight * loss_term

            for sfx in ("", "-original"):
                weight = self.bottom_mask
                if sfx != "":
                    weight *= self.auxiliary

                with torch.no_grad():
                    loss_inclusion_mask = (target["is_bad_labels"] < 1e-7).to(
                        torch.float32
                    )
                    # Check whether line is masked out
                    if self.ignore_lines_during_passive:
                        loss_inclusion_mask *= 1 - target["is_passive"]
                    if self.ignore_lines_during_removed:
                        loss_inclusion_mask *= 1 - target["is_removed"]

                if not weight:
                    continue
                elif "logit_is_boundary_bottom" + sfx in input:
                    # Load cross-entropy class target
                    C = target["index_bottom" + sfx].to(
                        device=input["logit_is_boundary_bottom" + sfx + cs].device,
                        dtype=torch.long,
                    )
                    loss_term = F.cross_entropy(
                        input["logit_is_boundary_bottom" + sfx + cs].transpose(-2, -1),
                        C,
                        reduction="none",
                    )
                    loss_term *= cmask.unsqueeze(-1)
                    loss_term *= loss_inclusion_mask
                    if self.reduction == "mean":
                        loss_term = torch.mean(loss_term)
                    elif self.reduction == "sum":
                        loss_term = torch.sum(loss_term)
                    elif self.reduction != "none":
                        raise ValueError(
                            "Unsupported reduction: {}".format(self.reduction)
                        )
                elif "logit_is_below_bottom" + sfx in input:
                    warnings.warn(
                        'Using loss corresponding to "mask" logits.'
                        ' The "boundary" is recommended instead.'
                        " The loss component for this line will be"
                        " F.binary_cross_entropy_with_logits(input[{}], target[{}])"
                        "".format("logit_is_below_bottom" + sfx + cs, target_key)
                    )
                    loss_term = F.binary_cross_entropy_with_logits(
                        input["logit_is_below_bottom" + sfx + cs],
                        target["mask_bottom" + sfx].to(
                            input["logit_is_below_bottom" + sfx + cs].device,
                            input["logit_is_below_bottom" + sfx + cs].dtype,
                        ),
                        reduction="none",
                    )
                    loss_term *= cmask.unsqueeze(-1).unsqueeze(-1)
                    loss_term *= loss_inclusion_mask.unsqueeze(-1)
                    if self.reduction == "mean":
                        loss_term = torch.mean(loss_term)
                    elif self.reduction == "sum":
                        loss_term = torch.sum(loss_term)
                    elif self.reduction != "none":
                        raise ValueError(
                            "Unsupported reduction: {}".format(self.reduction)
                        )
                else:
                    raise ValueError(
                        "The input does not contain either {} or {} fields."
                        " At least one of these is required if the loss term weighting"
                        " is non-zero.".format(
                            "logit_is_boundary_bottom" + sfx,
                            "logit_is_below_bottom" + sfx,
                        )
                    )
                if torch.isnan(loss_term).any():
                    print("Loss term mask_bottom{} is NaN".format(sfx))
                else:
                    closs += weight * loss_term

            if self.removed_segment:
                loss_term = F.binary_cross_entropy_with_logits(
                    input["logit_is_removed" + cs],
                    target["is_removed"].to(
                        input["logit_is_removed" + cs].device,
                        input["logit_is_removed" + cs].dtype,
                    ),
                    reduction="none",
                )
                loss_term *= cmask.unsqueeze(-1)
                if self.reduction == "mean":
                    loss_term = torch.mean(loss_term)
                elif self.reduction == "sum":
                    loss_term = torch.sum(loss_term)
                elif self.reduction != "none":
                    raise ValueError("Unsupported reduction: {}".format(self.reduction))
                if torch.isnan(loss_term).any():
                    print("Loss term is_removed is NaN")
                else:
                    closs += self.removed_segment * loss_term

            if self.passive:
                loss_term = self.passive * F.binary_cross_entropy_with_logits(
                    input["logit_is_passive" + cs],
                    target["is_passive"].to(
                        input["logit_is_passive" + cs].device,
                        input["logit_is_passive" + cs].dtype,
                    ),
                    reduction="none",
                )
                loss_term *= cmask.unsqueeze(-1)
                if self.reduction == "mean":
                    loss_term = torch.mean(loss_term)
                elif self.reduction == "sum":
                    loss_term = torch.sum(loss_term)
                elif self.reduction != "none":
                    raise ValueError("Unsupported reduction: {}".format(self.reduction))
                if torch.isnan(loss_term).any():
                    print("Loss term is_passive is NaN")
                else:
                    closs += self.passive * loss_term

            for sfx in ("", "-original", "-ntob"):
                weight = self.patch
                if sfx != "":
                    weight *= self.auxiliary
                if not weight:
                    continue
                loss_term = F.binary_cross_entropy_with_logits(
                    input["logit_is_patch" + sfx + cs],
                    target["mask_patches" + sfx].to(
                        input["logit_is_patch" + sfx + cs].device,
                        input["logit_is_patch" + sfx + cs].dtype,
                    ),
                    reduction="none",
                )
                loss_term *= cmask.unsqueeze(-1).unsqueeze(-1)
                if self.reduction == "mean":
                    loss_term = torch.mean(loss_term)
                elif self.reduction == "sum":
                    loss_term = torch.sum(loss_term)
                elif self.reduction != "none":
                    raise ValueError("Unsupported reduction: {}".format(self.reduction))
                if torch.isnan(loss_term).any():
                    print("Loss term mask_patches{} is NaN".format(sfx))
                else:
                    closs += weight * loss_term

            if self.overall:
                loss_term = self.overall * F.binary_cross_entropy(
                    input["p_keep_pixel" + cs],
                    target["mask"].to(
                        input["p_keep_pixel" + cs].device,
                        input["p_keep_pixel" + cs].dtype,
                    ),
                    reduction="none",
                )
                loss_term *= cmask.unsqueeze(-1).unsqueeze(-1)
                if self.reduction == "mean":
                    loss_term = torch.mean(loss_term)
                elif self.reduction == "sum":
                    loss_term = torch.sum(loss_term)
                elif self.reduction != "none":
                    raise ValueError("Unsupported reduction: {}".format(self.reduction))
                if torch.isnan(loss_term).any():
                    print("Loss term overall is NaN")
                else:
                    closs += self.overall * loss_term

            loss += closs

        # Avoid double-counting the loss
        if n_conditions_in_loss > 1:
            loss /= 2

        return loss



"""
Meters for tracking measurements during training.
"""

# This file is part of Echofilter.
#
# Copyright (C) 2020-2022  Scott C. Lowe and Offshore Energy Research Association (OERA)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import torch


class AverageMeter(object):
    """Compute and store the average and current value."""

    def __init__(self, name, fmt=":f"):
        self.name = name
        self.fmt = fmt
        self.reset()

    def reset(self):
        self.values = []
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=None):
        if isinstance(val, torch.Tensor):
            val = val.cpu().numpy()
        if isinstance(val, (int, float)):
            values = [val]
        else:
            values = list(val)
            val = np.mean(val)
        if n is None:
            n = len(values)
        self.values += values
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

    def __str__(self):
        fmtstr = "{name} {val" + self.fmt + "} ({avg" + self.fmt + "})"
        return fmtstr.format(**self.__dict__)


class ProgressMeter(object):
    def __init__(self, num_batches, meters, prefix=""):
        self.batch_fmtstr = self._get_batch_fmtstr(num_batches)
        self.meters = meters
        self.prefix = prefix

    def display(self, batch):
        entries = [self.prefix + self.batch_fmtstr.format(batch)]
        entries += [str(meter) for meter in self.meters]
        print("    ".join(entries))

    def _get_batch_fmtstr(self, num_batches):
        num_digits = len(str(num_batches // 1))
        fmt = "{:" + str(num_digits) + "d}"
        return "[" + fmt + "/" + fmt.format(num_batches) + "]"



"""
Dataset metadata, relevant for loading correct data.
"""

# This file is part of Echofilter.
#
# Copyright (C) 2020-2022  Scott C. Lowe and Offshore Energy Research Association (OERA)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os

import numpy as np


def recall_passive_edges(sample_path, timestamps):
    """
    Define passive data edges for samples within known datasets.

    Parameters
    ----------
    sample_path : str
        Path to sample.
    timestamps : array_like vector
        Vector of timestamps in sample.

    Returns
    -------
    passive_starts : numpy.ndarray or None
        Indices indicating the onset of passive data collection periods, or
        ``None`` if passive metadata is unavailable for this sample.
    passive_ends : numpy.ndarray or None
        Indices indicating the offset of passive data collection periods, or
        ``None`` if passive metadata is unavailable for this sample.
    finder_version : absent or str
        If ``passive_starts`` and ``passive_ends``, this string may be present to
        indicate which passive finder algorithm works best for this dataset.
    """
    sample_path = sample_path.lower()
    sample_parts = os.path.normpath(sample_path).split(os.path.sep)

    nt = len(timestamps)

    if "minaspassage" in sample_parts:
        if "december2017" in sample_path:
            return np.array([]), np.array([])
        elif "march2018" in sample_path:
            # No clear best detector, but all samples well behaved
            passive_starts = np.arange(0, nt, 360)
            passive_ends = passive_starts + 60
            return passive_starts, passive_ends
        elif "september2018" in sample_path:
            # Some samples are not as expected
            for name in [
                "D20181021-T165220_D20181021-T222221",
                "D20181022-T105220_D20181022-T162217",
                "D20181022-T172213_D20181022-T232217",
                "D20181026-T082220_D20181026-T135213",
                "D20181026-T142217_D20181026-T195218",
            ]:
                # Detector v1 finds the correct passive boundaries, so we leave
                # it to that to work it out
                if name in sample_path:
                    return None, None, "v1"
            passive_starts = np.arange(300, nt, 360)
            passive_ends = passive_starts + 60
            return passive_starts, passive_ends

    elif "grandpassage" in sample_parts:
        like_phase1 = False
        if "phase2" in sample_parts:
            # Some samples have the same passive data as in phase1
            for name in [
                "WBAT_2B_20200125_UTC100017_floodhigh",
                "WBAT_2B_20200125_UTC160020_ebblow",
                "WBAT_2B_20200127_UTC000020_floodhigh",
                "WBAT_2B_20200127_UTC060020_ebblow",
                "WBAT_2B_20200127_UTC120021_floodhigh",
                "WBAT_2B_20200127_UTC180020_ebblow",
                "WBAT_2B_20200128_UTC000017_floodhigh",
                "WBAT_2B_20200128_UTC060017_ebblow",
                "WBAT_2B_20200128_UTC120023_floodhigh",
                "WBAT_2B_20200128_UTC180017_ebblow",
                "WBAT_2B_20200130_UTC080017_ebblow",
                "WBAT_2B_20200130_UTC140020_floodhigh",
                "WBAT_2B_20200130_UTC200017_ebblow",
                "WBAT_2B_20200131_UTC020020_floodhigh",
                "WBAT_2B_20200131_UTC080021_ebblow",
                "WBAT_2B_20200131_UTC140022_floodhigh",
                "WBAT_2B_20200202_UTC040019_floodhigh",
                "WBAT_2B_20200202_UTC100022_ebblow",
                # "WBAT_2B_20200130_UTC020017_floodhigh",  # broken csv encoding
            ]:
                if name in sample_path:
                    like_phase1 = True
            # There are 5 other passive timing schedules across the other
            # 14 samples. The passive data finder works well for all of these.
            if not like_phase1:
                return None, None, "v1"

        if like_phase1 or "phase1" in sample_parts:
            passive_starts = np.arange(-200, nt, 3420)
            passive_ends = passive_starts + 420
            passive_starts = np.maximum(passive_starts, 0)
            passive_ends = np.minimum(passive_ends, nt)

    # For the mobile dataset, v1 and v2 always agree and appear to be 100%
    # accurate.

    return None, None



"""
Connectors and pathing modules.
"""

import torch
import torch.nn.functional as F
from torch import nn

from .conv import PointwiseConv2d


class ResidualConnect(nn.Module):
    """
    Joins up a residual connection, correcting for changes in number of channels.
    """

    def __init__(self, in_channels, out_channels):
        super(ResidualConnect, self).__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels

        if in_channels == out_channels:
            # No need to do anything to our passthru input
            self.op = nn.Identity()
        elif in_channels < out_channels:
            # The number of channels has increased, so keep the original
            # input as it is, and pad up to match the size of the "residual".
            self.op = PointwiseConv2d(in_channels, out_channels - in_channels)
        else:
            # The number of channels has decreased, so we need to map the
            # original down to the size of the "residual".
            self.op = PointwiseConv2d(in_channels, out_channels)

    def forward(self, residual, passed_thru):
        if self.in_channels < self.out_channels:
            return residual + torch.cat([passed_thru, self.op(passed_thru)], dim=1)
        return residual + self.op(passed_thru)


class FlexibleConcat2d(nn.Module):
    """
    Concatenate two inputs of nearly the same shape.
    """

    def forward(self, x1, x2):
        """
        Forward step.

        Parameters
        ----------
        x1 : torch.Tensor
            Tensor, possibly smaller than ``x2``.
        x2 : torch.Tensor
            Tensor, at least as large as ``x1``.

        Returns
        -------
        torch.Tensor
            Concatenated ``x1`` (padded if necessary) and ``x2``, along
            dimension ``1``.
        """
        # input is CHW
        diffY = torch.tensor([x2.shape[-2] - x1.shape[-2]])
        diffX = torch.tensor([x2.shape[-1] - x1.shape[-1]])

        if diffX != 0 or diffY != 0:
            x1 = F.pad(
                x1,
                [diffX // 2, diffX - diffX // 2, diffY // 2, diffY - diffY // 2],
                mode="replicate",
            )
        # if you have padding issues, see
        # https://github.com/HaiyongJiang/U-Net-Pytorch-Unstructured-Buggy/commit/0e854509c2cea854e247a9c615f175f76fbb2e3a
        # https://github.com/xiaopeng-liao/Pytorch-UNet/commit/8ebac70e633bac59fc22bb5195e513d5832fb3bd
        return torch.cat([x1, x2], dim=1)



#!/usr/bin/env python

"""
Model training routine.
"""

# This file is part of Echofilter.
#
# Copyright (C) 2020-2022  Scott C. Lowe and Offshore Energy Research Association (OERA)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import copy
import datetime
import os
import pprint
import shutil
import sys
import time
import traceback
from collections import OrderedDict

try:
    import apex
except ImportError:
    apex = None

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import torch
import torch.nn
import torch.optim
import torch.utils.data
import torchvision.transforms

import echofilter.data
import echofilter.optim.utils
import echofilter.raw.shardloader
from echofilter.nn.unet import UNet
from echofilter.nn.utils import count_parameters, seed_all
from echofilter.nn.wrapper import Echofilter, EchofilterLoss
from echofilter.optim import criterions, schedulers
from echofilter.optim.meters import AverageMeter, ProgressMeter
from echofilter.plotting import plot_transect_predictions
from echofilter.raw.loader import get_partition_list
from echofilter.raw.manipulate import load_decomposed_transect_mask
from echofilter.ui.train_cli import main

# --- For mobile dataset,
# DATA_CENTER = -81.5
# DATA_DEVIATION = 21.9
# CENTER_METHOD = "mean"
# DEVIATION_METHOD = "stdev"

# --- For stationary dataset,
# DATA_CENTER = -78.7
# DATA_DEVIATION = 19.2
# CENTER_METHOD = "mean"
# DEVIATION_METHOD = "stdev"

# --- For intermediate values between both datasets
# DATA_CENTER = -80.
# DATA_DEVIATION = 20.
# CENTER_METHOD = "mean"
# DEVIATION_METHOD = "stdev"
# NAN_VALUE = -3

# --- Overall values to use
# DATA_CENTER = -97.5
# DATA_DEVIATION = 16.5
# CENTER_METHOD = "pc10"
# DEVIATION_METHOD = "idr"
# NAN_VALUE = -1

# Normalise each sample independently, based on its own intensity distribution
CENTER_METHOD = "median"
DEVIATION_METHOD = "idr"
DATA_CENTER = CENTER_METHOD
DATA_DEVIATION = DEVIATION_METHOD
NAN_VALUE = -3

# Transects to plot for debugging
PLOT_TRANSECTS = {
    "mobile": [
        "mobile/Survey07/Survey07_GR4_N5W_survey7",
        "mobile/Survey14/Survey14_GR4_N0W_E",
        "mobile/Survey16/Survey16_GR4_N5W_E",
        "mobile/Survey17/Survey17_GR4_N5W_E",
    ],
    "MinasPassage": [
        "MinasPassage/december2017/december2017_D20180213-T115216_D20180213-T172216",
        "MinasPassage/march2018/march2018_D20180513-T195216_D20180514-T012216",
        "MinasPassage/september2018/september2018_D20181027-T202217_D20181028-T015217",
        "MinasPassage/september2018/september2018_D20181107-T122220_D20181107-T175217",
    ],
    "GrandPassage": [
        "GrandPassage/phase2/GrandPassage_WBAT_2B_20200125_UTC160020_ebblow",
        "GrandPassage/phase2/GrandPassage_WBAT_2B_20200202_UTC040019_floodhigh",
    ],
}

DEFAULT_CROP_DEPTH_PLOTS = 100
MAX_INPUT_LEN = 3500


def train(
    data_dir="/data/dsforce/surveyExports",
    dataset_name="mobile",
    train_partition=None,
    val_partition=None,
    sample_shape=(128, 512),
    crop_depth=None,
    resume="",
    restart="",
    log_name=None,
    log_name_append=None,
    conditional=False,
    n_block=6,
    latent_channels=32,
    expansion_factor=1,
    expand_only_on_down=False,
    blocks_per_downsample=(2, 1),
    blocks_before_first_downsample=(2, 1),
    always_include_skip_connection=True,
    deepest_inner="horizontal_block",
    intrablock_expansion=6,
    se_reduction=4,
    downsampling_modes="max",
    upsampling_modes="bilinear",
    depthwise_separable_conv=True,
    residual=True,
    actfn="InplaceReLU",
    kernel_size=5,
    use_mixed_precision=None,
    amp_opt="O1",
    device="cuda",
    multigpu=False,
    n_worker=8,
    batch_size=16,
    stratify=True,
    n_epoch=20,
    seed=None,
    print_freq=50,
    optimizer="adam",
    schedule="constant",
    lr=0.1,
    momentum=0.9,
    base_momentum=None,
    weight_decay=1e-5,
    warmup_pct=0.2,
    warmdown_pct=0.7,
    anneal_strategy="cos",
    overall_loss_weight=0.0,
):
    """
    Train a model.
    """
    if restart and not resume:
        raise ValueError(
            "A checkpoint must be provided to restart from when doing a cold restart"
        )

    # Lazy import of tensorboard, so training-only requirements are not needed
    # for automated documentation building.
    from torch.utils.tensorboard import SummaryWriter

    seed_all(seed)
    # Can't get this to be deterministic anyway, so may as well keep the
    # non-deterministic optimization enabled
    torch.backends.cudnn.deterministic = False
    torch.backends.cudnn.benchmark = True

    # Input handling
    schedule = schedule.lower()

    if base_momentum is None:
        base_momentum = momentum

    if log_name is None or log_name == "":
        log_name = datetime.datetime.now().strftime("%Y-%m-%d_%H.%M.%S")
    if log_name_append is None:
        log_name_append = os.uname()[1]
    if len(log_name_append) > 0:
        log_name += "_" + log_name_append

    print("Output will be written to {}/{}".format(dataset_name, log_name))

    if use_mixed_precision is None:
        use_mixed_precision = "cpu" not in device
    if use_mixed_precision and apex is None:
        print("NVIDIA apex must be installed to use mixed precision.")
        use_mixed_precision = False

    # Need to set the default device for apex.amp
    if device is not None and device != "cpu":
        torch.cuda.set_device(torch.device(device))

    # Build dataset
    dataset_train, dataset_val, dataset_augval = build_dataset(
        dataset_name,
        data_dir,
        sample_shape,
        train_partition=train_partition,
        val_partition=val_partition,
    )

    print("Train dataset has {:4d} samples".format(len(dataset_train)))
    print("Val   dataset has {:4d} samples".format(len(dataset_val)))

    stratify_allowed = stratify
    if stratify_allowed:
        stratify = isinstance(dataset_train, echofilter.data.dataset.ConcatDataset)
    if stratify:
        # Use custom stratified sampler to handle multiple datasets
        sampler = echofilter.data.dataset.StratifiedRandomSampler(dataset_train)
    else:
        sampler = None

    loader_train = torch.utils.data.DataLoader(
        dataset_train,
        batch_size=batch_size,
        shuffle=not stratify,
        sampler=sampler,
        num_workers=n_worker,
        pin_memory=True,
        drop_last=True,
        worker_init_fn=echofilter.data.utils.worker_seed_fn,
    )
    loader_val = torch.utils.data.DataLoader(
        dataset_val,
        batch_size=batch_size,
        shuffle=False,
        num_workers=n_worker,
        pin_memory=True,
        drop_last=False,
        worker_init_fn=echofilter.data.utils.worker_staticseed_fn,
    )
    loader_augval = torch.utils.data.DataLoader(
        dataset_augval,
        batch_size=batch_size,
        shuffle=False,
        num_workers=n_worker,
        pin_memory=True,
        drop_last=False,
        worker_init_fn=echofilter.data.utils.worker_staticseed_fn,
    )
    print("Train loader has {:3d} batches".format(len(loader_train)))
    print("Val   loader has {:3d} batches".format(len(loader_val)))

    print()
    print(
        "Constructing U-Net model with "
        "{} blocks, "
        "initial latent channels {}, "
        "expansion_factor {}".format(n_block, latent_channels, expansion_factor)
    )
    model_parameters = dict(
        in_channels=1,
        out_channels=10,
        initial_channels=latent_channels,
        bottleneck_channels=latent_channels,
        n_block=n_block,
        unet_expansion_factor=expansion_factor,
        expand_only_on_down=expand_only_on_down,
        blocks_per_downsample=blocks_per_downsample,
        blocks_before_first_downsample=blocks_before_first_downsample,
        always_include_skip_connection=always_include_skip_connection,
        deepest_inner=deepest_inner,
        intrablock_expansion=intrablock_expansion,
        se_reduction=se_reduction,
        downsampling_modes=downsampling_modes,
        upsampling_modes=upsampling_modes,
        depthwise_separable_conv=depthwise_separable_conv,
        residual=residual,
        actfn=actfn,
        kernel_size=kernel_size,
    )
    if conditional:
        model_parameters["out_channels"] *= 3
    print()
    pprint.pprint(model_parameters)
    print()

    unet = UNet(**model_parameters)
    unet.to(device)
    print(
        "Built UNet model with {} trainable parameters".format(
            count_parameters(unet, only_trainable=True)
        )
    )

    # define loss function (criterion) and optimizer
    criterion = EchofilterLoss(conditional=conditional, overall=overall_loss_weight)

    optimizer_name = optimizer.lower()
    if optimizer_name == "adam":
        optimizer_class = torch.optim.Adam
    elif optimizer_name == "adamw":
        optimizer_class = torch.optim.AdamW
    elif optimizer_name == "ranger":
        import ranger

        optimizer_class = ranger.Ranger
    elif optimizer_name == "rangerva":
        import ranger

        optimizer_class = ranger.RangerVA
    elif optimizer_name == "rangerqh":
        import ranger

        optimizer_class = ranger.RangerQH
    else:
        # We don't support arbitrary optimizers from torch.optim because they
        # need different configuration parameters to Adam.
        raise ValueError("Unrecognised optimizer: {}".format(optimizer))

    optimizer = optimizer_class(
        unet.parameters(),
        lr,
        betas=(momentum, 0.999),
        weight_decay=weight_decay,
    )

    schedule_data = {"name": schedule}
    lr_initial_div_factor = 1e3
    lr_final_div_factor = 1e5
    if schedule == "lrfinder":
        pass
    elif schedule == "constant":
        pass
    elif schedule == "onecycle":
        schedule_data["scheduler"] = schedulers.OneCycleLR(
            optimizer,
            max_lr=lr,
            steps_per_epoch=len(loader_train),
            epochs=n_epoch,
            pct_start=warmup_pct,
            anneal_strategy=anneal_strategy,
            cycle_momentum=True,
            base_momentum=base_momentum,
            max_momentum=momentum,
            div_factor=lr_initial_div_factor,
            final_div_factor=lr_final_div_factor,
        )
    elif schedule == "mesaonecycle":
        schedule_data["scheduler"] = schedulers.MesaOneCycleLR(
            optimizer,
            max_lr=lr,
            steps_per_epoch=len(loader_train),
            epochs=n_epoch,
            pct_start=warmup_pct,
            pct_end=warmdown_pct,
            anneal_strategy=anneal_strategy,
            cycle_momentum=True,
            base_momentum=base_momentum,
            max_momentum=momentum,
            div_factor=lr_initial_div_factor,
            final_div_factor=lr_final_div_factor,
        )
    else:
        raise ValueError("Unsupported schedule: {}".format(schedule))

    if use_mixed_precision:
        print('Converting unet to mixed precision, opt="{}"'.format(amp_opt))
        unet, optimizer = apex.amp.initialize(unet, optimizer, opt_level=amp_opt)

    model_inner = unet
    if multigpu and torch.cuda.device_count() > 1:
        print("Using local parallelism on {} GPUs".format(torch.cuda.device_count()))
        model_inner = torch.nn.DataParallel(model_inner)

    # Add UI wrapper around model
    model = Echofilter(
        model_inner,
        top="boundary",
        bottom="boundary",
        conditional=conditional,
    )

    if schedule == "lrfinder":
        from torch_lr_finder import LRFinder

        print("Running learning rate finder")
        lr_finder = LRFinder(model, optimizer, criterion, device=device)
        lr_finder.range_test(loader_train, end_lr=100, num_iter=100, diverge_th=3)
        print("Plotting learning rate finder results")
        hf = plt.figure(figsize=(15, 9))
        ax = plt.axes()
        lr_finder.plot(skip_start=0, skip_end=1, log_lr=True, ax=ax)
        plt.tick_params(reset=True, color=(0.2, 0.2, 0.2))
        plt.tick_params(labelsize=14)
        ax.minorticks_on()
        ax.tick_params(direction="out")
        # Save figure
        figpth = os.path.join("models", dataset_name, log_name, "lrfinder.png")
        os.makedirs(os.path.dirname(figpth), exist_ok=True)
        plt.savefig(figpth)
        print("LR Finder results saved to {}".format(figpth))
        return

    # Initialise loop tracking
    start_epoch = 1
    best_loss_val = float("inf")

    # optionally resume from a checkpoint
    if resume:
        if not os.path.isfile(resume):
            raise EnvironmentError("No checkpoint found at '{}'".format(resume))
        print("Loading checkpoint '{}'".format(resume))
        if device is None:
            checkpoint = torch.load(resume)
        else:
            # Map model to be loaded to specified single gpu.
            checkpoint = torch.load(resume, map_location=device)
        try:
            unet.load_state_dict(checkpoint["state_dict"])
        except RuntimeError:
            print(
                "Checkpoint doesn't seem to be for the UNet component. Trying"
                " to load it as the whole model instead."
            )
            model.load_state_dict(checkpoint["state_dict"])
        if restart == "cold":
            print("Loaded checkpoint '{}' for cold restart".format(resume))
        else:
            optimizer.load_state_dict(checkpoint["optimizer"])
            if use_mixed_precision and "amp" in checkpoint:
                apex.amp.load_state_dict(checkpoint["amp"])
        if restart == "warm":
            # We loaded the optimizer state from the checkpoint to get the
            # current buffer values (the weighted history of updates).
            # However, this also changed our schedule settings to the ones
            # used to train the model in the checkpoint, so we need to
            # overwrite it with the schedule initialisation step now.
            for group in optimizer.param_groups:
                if "onecycle" in schedule:
                    group["lr"] = group["initial_lr"] = lr / lr_initial_div_factor
                    group["max_lr"] = lr
                    group["min_lr"] = group["initial_lr"] / lr_final_div_factor
                    group["base_momentum"] = base_momentum
                    group["max_momentum"] = momentum
                else:
                    group["lr"] = lr
                    if "betas" in group:
                        _, beta2 = group["betas"]
                        group["betas"] = (momentum, beta2)
                    else:
                        group["momentum"] = momentum
            print("Loaded checkpoint '{}' for warm restart".format(resume))
        elif not restart:
            start_epoch = checkpoint["epoch"] + 1
            best_loss_val = checkpoint["best_loss"]
            if "scheduler" in schedule_data:
                step_num = checkpoint["epoch"] * len(loader_train)
                schedule_data["scheduler"].last_epoch = step_num
            print(
                "Loaded checkpoint '{}' (epoch {})".format(resume, checkpoint["epoch"])
            )

    # Make a tensorboard writer
    writer = SummaryWriter(log_dir=os.path.join("runs", dataset_name, log_name))

    print("Starting training")
    t_start = time.time()
    for epoch in range(start_epoch, n_epoch + 1):

        t_epoch_start = time.time()

        # Set the seed state at the start of each epoch
        # This ensures some level of consistency between models which continue
        # running and those resumed from checkpoints (don't want to run the
        # exact initial augmented samples from the first epoch again when
        # resuming later in the run)
        seed_all(1000 + seed + epoch)
        # Can't get this to be deterministic anyway, so may as well keep the
        # non-deterministic optimization enabled
        torch.backends.cudnn.deterministic = False
        torch.backends.cudnn.benchmark = True

        # Resample offsets for each window
        loader_train.dataset.initialise_datapoints()

        # train for one epoch
        (
            loss_tr,
            meters_tr,
            (ex_input_tr, ex_data_tr, ex_output_tr),
            (batch_time, data_time),
        ) = train_epoch(
            loader_train,
            model,
            criterion,
            optimizer,
            device,
            epoch,
            print_freq=print_freq,
            schedule_data=schedule_data,
            use_mixed_precision=use_mixed_precision,
        )

        t_val_start = time.time()

        # evaluate on validation set
        loss_val, meters_val, (ex_input_val, ex_data_val, ex_output_val) = validate(
            loader_val,
            model,
            criterion,
            device,
            print_freq=print_freq,
            prefix="Validation",
        )
        # evaluate on augmented validation set
        (
            loss_augval,
            meters_augval,
            (ex_input_augval, ex_data_augval, ex_output_augval),
        ) = validate(
            loader_augval,
            model,
            criterion,
            device,
            print_freq=print_freq,
            prefix="Aug-Val   ",
        )
        t_val_end = time.time()

        print(
            "{}/{}\nCompleted {} of {} epochs in {}".format(
                dataset_name,
                log_name,
                epoch,
                n_epoch,
                datetime.timedelta(seconds=time.time() - t_start),
            )
        )
        # Print metrics to terminal
        name_fmt = "{:.<28s}"
        current_lr = echofilter.optim.utils.get_current_lr(optimizer)
        current_mom = echofilter.optim.utils.get_current_momentum(optimizer)

        print((name_fmt + " {:.4e}").format("Learning rate", current_lr))
        print((name_fmt + " {:.4f}").format("Momentum", current_mom))
        print(
            (name_fmt + " Train: {:.4e}  AugVal: {:.4e}  Val: {:.4e}").format(
                "Loss", loss_tr, loss_augval, loss_val
            )
        )
        for chn in meters_tr:
            if chn.lower() != "overall":
                continue
            # For each output plane
            print(chn)
            for cr in meters_tr[chn]:
                # For each criterion
                fmt_str = name_fmt
                fmt_str += " Train: {" + meters_tr[chn][cr].fmt + "}"
                fmt_str += "    AugVal: {" + meters_augval[chn][cr].fmt + "}"
                fmt_str += "    Val: {" + meters_val[chn][cr].fmt + "}"
                print(
                    fmt_str.format(
                        meters_tr[chn][cr].name,
                        meters_tr[chn][cr].avg,
                        meters_augval[chn][cr].avg,
                        meters_val[chn][cr].avg,
                    )
                )

        # Add hyper parameters to tensorboard
        writer.add_scalar("learning_rate", current_lr, epoch)
        writer.add_scalar("momentum", current_mom, epoch)
        writer.add_scalar(
            "parameter_count", count_parameters(model, only_trainable=True), epoch
        )

        # Add metrics to tensorboard
        for loss_p, partition in (
            (loss_tr, "Train"),
            (loss_val, "Val"),
            (loss_augval, "ValAug"),
        ):
            writer.add_scalar("{}/{}".format("Loss", partition), loss_p, epoch)
        for chn in meters_tr:
            # For each output plane
            for cr in meters_tr[chn]:
                # For each criterion
                writer.add_scalar(
                    "{}/{}/{}".format(cr, chn, "Train"), meters_tr[chn][cr].avg, epoch
                )
                writer.add_scalar(
                    "{}/{}/{}".format(cr, chn, "ValAug"),
                    meters_augval[chn][cr].avg,
                    epoch,
                )
                writer.add_scalar(
                    "{}/{}/{}".format(cr, chn, "Val"), meters_val[chn][cr].avg, epoch
                )

        # Determine whether to generate sample transect plots, or skip them
        if n_epoch < 20:
            # Every epoch
            generate_sample_images = True
        else:
            # Every 20th of the way through training
            generate_sample_images = (
                int(20 * epoch / n_epoch) > 20 * (epoch - 1) / n_epoch
            )
        # But always generate samples for first two epochs and last epoch
        if epoch <= 2 or epoch == n_epoch:
            generate_sample_images = True

        def ensure_clim_met(x, x0=0.0, x1=1.0):
            x = x.clone()
            x[0, :, 0, 0] = 0
            x[0, :, 0, 1] = 1
            return x

        def add_image_border(x):
            """
            Add a green border around a a tensor of images.

            Parameters
            ----------
            x : torch.Tensor
                Tensor in NCWH or NCHW format.

            Returns
            -------
            torch.Tensor
                As ``x``, but padded with a green border.
            """
            if x.shape[1] == 1:
                x = torch.cat([x, x, x], dim=1)
            if x.shape[1] != 3:
                raise ValueError("RGB image needs three color channels")
            shp = list(x.shape)
            shp[-1] = 1
            x = torch.cat(
                [
                    torch.zeros(shp, dtype=x.dtype, device=x.device),
                    x,
                    torch.zeros(shp, dtype=x.dtype, device=x.device),
                ],
                dim=-1,
            )
            shp = list(x.shape)
            shp[-2] = 1
            x = torch.cat(
                [
                    torch.zeros(shp, dtype=x.dtype, device=x.device),
                    x,
                    torch.zeros(shp, dtype=x.dtype, device=x.device),
                ],
                dim=-2,
            )
            x[:, 1, :, 0] = 1.0
            x[:, 1, :, -1] = 1.0
            x[:, 1, 0, :] = 1.0
            x[:, 1, -1, :] = 1.0
            return x

        # Add example images to tensorboard
        for (ex_input, ex_data, ex_output), partition in (
            ((ex_input_tr, ex_data_tr, ex_output_tr), "Train"),
            ((ex_input_val, ex_data_val, ex_output_val), "Val"),
            ((ex_input_augval, ex_data_augval, ex_output_augval), "ValAug"),
        ):
            if not generate_sample_images:
                continue
            writer.add_images(
                "Input/" + partition,
                ex_input,
                epoch,
                dataformats="NCWH",
            )
            writer.add_images(
                "Overall/" + partition + "/Target",
                ensure_clim_met(add_image_border(ex_data["mask"].float().unsqueeze(1))),
                epoch,
                dataformats="NCWH",
            )
            writer.add_images(
                "Overall/" + partition + "/Output/p",
                ensure_clim_met(
                    add_image_border(ex_output["p_keep_pixel"].unsqueeze(1))
                ),
                epoch,
                dataformats="NCWH",
            )
            writer.add_images(
                "Overall/" + partition + "/Overlap",
                ensure_clim_met(
                    add_image_border(
                        torch.stack(
                            [
                                ex_output["mask_keep_pixel"].float(),
                                torch.zeros_like(
                                    ex_output["mask_keep_pixel"], dtype=torch.float
                                ),
                                ex_data["mask"].float(),
                            ],
                            dim=1,
                        )
                    )
                ),
                epoch,
                dataformats="NCWH",
            )

        for k, plot_transects_k in PLOT_TRANSECTS.items():
            if not generate_sample_images:
                continue
            if "stationary" in dataset_name and k != "mobile":
                pass
            elif k not in dataset_name:
                continue
            plot_crop_depth = crop_depth
            if plot_crop_depth is None:
                plot_crop_depth = DEFAULT_CROP_DEPTH_PLOTS
            for transect_name in plot_transects_k:
                transect, prediction = generate_from_shards(
                    os.path.join(data_dir + "_sharded", transect_name),
                    model,
                    sample_shape=sample_shape,
                    crop_depth=plot_crop_depth,
                    device=device,
                    dtype=torch.float,
                )
                hf = plt.figure(figsize=(15, 9))
                plot_transect_predictions(
                    transect, prediction, cmap="viridis", linewidth=1
                )
                transect_name = transect_name.replace("/evExports", "")
                if epoch == n_epoch:
                    # Only save png if this is the final epoch
                    figpth = os.path.join(
                        "models",
                        dataset_name,
                        log_name,
                        "samples",
                        transect_name + "_output.png",
                    )
                    os.makedirs(os.path.dirname(figpth), exist_ok=True)
                    plt.savefig(figpth)
                writer.add_figure(transect_name, hf, epoch, close=True)

        # remember best loss and save checkpoint
        is_best = loss_val < best_loss_val
        best_loss_val = min(loss_val, best_loss_val)

        checkpoint = {
            "model_parameters": model_parameters,
            "sample_shape": sample_shape,
            "epoch": epoch,
            "state_dict": unet.state_dict(),
            "best_loss": best_loss_val,
            "optimizer": optimizer.state_dict(),
            "meters": meters_val,
            "data_center": DATA_CENTER,
            "data_deviation": DATA_DEVIATION,
            "center_method": CENTER_METHOD,
            "deviation_method": DEVIATION_METHOD,
            "nan_value": NAN_VALUE,
            "wrapper_mapping": model.mapping,
            "wrapper_params": model.params,
            "training_routine": "echofilter-train {}".format(echofilter.__version__),
        }
        if use_mixed_precision:
            checkpoint["amp"] = apex.amp.state_dict()

        if (
            schedule == "mesaonecycle"
            and epoch / n_epoch <= warmdown_pct
            and (epoch + 1) / n_epoch > warmdown_pct
        ):
            dup = "ep{}".format(epoch)
        else:
            dup = None
        save_checkpoint(
            checkpoint,
            is_best,
            dirname=os.path.join("models", dataset_name, log_name),
            dup=dup,
        )
        meters_to_csv(
            meters_val, is_best, dirname=os.path.join("models", dataset_name, log_name)
        )

        # Note how long everything took
        writer.add_scalar("time/batch", batch_time.avg, epoch)
        writer.add_scalar("time/batch/data", data_time.avg, epoch)
        writer.add_scalar("time/train", t_val_start - t_epoch_start, epoch)
        writer.add_scalar("time/val", t_val_end - t_val_start, epoch)
        writer.add_scalar("time/log", time.time() - t_val_end, epoch)
        writer.add_scalar("time/epoch", time.time() - t_epoch_start, epoch)

        # Ensure the tensorboard outputs for this epoch are flushed
        writer.flush()

    # Close tensorboard connection
    writer.close()


def build_dataset(
    dataset_name,
    data_dir,
    sample_shape,
    train_partition=None,
    val_partition=None,
    crop_depth=None,
    random_crop_args=None,
):
    """
    Construct a pytorch Dataset.

    Parameters
    ----------
    dataset_name : str
        Name of the dataset. This can optionally be a list of
        multiple datasets joined with ``"+"``.
    data_dir : str
        Path to root data directory, containing the dataset.
    sample_shape : iterable of length 2
        The shape which will be used for training.
    train_partition : str, optional
        Name of the partition to use for training. Can optionally be a list of
        multiple partitions joined with ``"+"``. Default is ``"train"``
        (except for ``stationary2`` where it is mixed).
    val_partition : str, optional
        Name of the partition to use for validation. Can optionally be a list
        of multiple partitions joined with ``"+"``. Default is ``"validate"``
        (except for ``stationary2`` where it is mixed).
    crop_depth : float or None, optional
        Depth at which to crop samples. Default is ``None``.
    random_crop_args : dict, optional
        Arguments to control the random crop used during training. Default is
        an empty dict, which uses the default arguments of
        :class`echofilter.data.transforms.RandomCropDepth`.

    Returns
    -------
    dataset_train : echofilter.data.dataset.TransectDataset
        Dataset of training samples.
    dataset_val : echofilter.data.dataset.TransectDataset
        Dataset of validation samples.
    dataset_augval : echofilter.data.dataset.TransectDataset
        Dataset of validation samples, appyling the training augmentation
        stack.
    """
    if random_crop_args is None:
        random_crop_args = {}

    if "+" in dataset_name:
        # Join multiple datasets together
        datasets = [
            build_dataset(
                subdataset_name,
                data_dir=data_dir,
                sample_shape=sample_shape,
                train_partition=train_partition,
                val_partition=val_partition,
                crop_depth=crop_depth,
                random_crop_args=random_crop_args,
            )
            for subdataset_name in dataset_name.split("+")
            if len(subdataset_name) > 0
        ]
        return tuple(
            echofilter.data.dataset.ConcatDataset([d[i] for d in datasets])
            for i in range(len(datasets[0]))
        )

    if dataset_name == "stationary2":
        # The stationary2 dataset is MinasPassage and GrandPassage,
        # plus a second duplicate copy of GrandPassage which uses only zoomed
        # out depth crops.
        random_crop_args2 = copy.deepcopy(random_crop_args)
        random_crop_args2["p_crop_is_none"] = 0.2
        random_crop_args2["p_crop_is_optimal"] = 0.0
        random_crop_args2["p_crop_is_close"] = 0.0
        # By default, we only evaluate on the validation partition of
        # MinasPassage, and train on the train partition of both plus the
        # validation partition of GrandPassage.
        if train_partition is None and val_partition is None:
            train_partition_main = "train"
            val_partition_main = "validate"
            train_partition_aux = "train+validate"
            val_partition_aux = ""
        else:
            train_partition_main = train_partition_aux = train_partition
            val_partition_main = val_partition_aux = val_partition
        # Assemble the datasets
        datasets = [
            build_dataset(
                "MinasPassage",
                data_dir=data_dir,
                sample_shape=sample_shape,
                train_partition=train_partition_main,
                val_partition=val_partition_main,
                crop_depth=crop_depth,
                random_crop_args=random_crop_args,
            ),
            build_dataset(
                "GrandPassage",
                data_dir=data_dir,
                sample_shape=sample_shape,
                train_partition=train_partition_aux,
                val_partition=val_partition_aux,
                crop_depth=crop_depth,
                random_crop_args=random_crop_args,
            ),
            build_dataset(
                "GrandPassage",
                data_dir=data_dir,
                sample_shape=sample_shape,
                train_partition=train_partition_aux,
                val_partition=val_partition_aux,
                crop_depth=crop_depth,
                random_crop_args=random_crop_args2,
            ),
        ]
        return tuple(
            echofilter.data.dataset.ConcatDataset([d[i] for d in datasets])
            for i in range(len(datasets[0]))
        )

    if train_partition is None:
        train_partition = "train"
    if val_partition is None:
        val_partition = "validate"

    # Augmentations
    train_transform = torchvision.transforms.Compose(
        [
            echofilter.data.transforms.RandomCropDepth(**random_crop_args),
            echofilter.data.transforms.RandomReflection(),
            echofilter.data.transforms.Normalize(DATA_CENTER, DATA_DEVIATION),
            echofilter.data.transforms.ColorJitter(0.5, 0.3),
            echofilter.data.transforms.ReplaceNan(NAN_VALUE),
            echofilter.data.transforms.RandomElasticGrid(
                sample_shape,
                order=None,
                p=0.5,
                sigma=[8, 16],
                alpha=0.1,
            ),
            echofilter.data.transforms.Rescale(sample_shape, order=None),
        ]
    )
    val_transform = torchvision.transforms.Compose(
        [
            echofilter.data.transforms.OptimalCropDepth(),
            echofilter.data.transforms.Normalize(DATA_CENTER, DATA_DEVIATION),
            echofilter.data.transforms.ReplaceNan(NAN_VALUE),
            echofilter.data.transforms.Rescale(sample_shape, order=1),
        ]
    )

    train_paths = []
    for partition_name in train_partition.split("+"):
        if len(partition_name) == 0:
            continue
        train_paths += get_partition_list(
            partition_name,
            dataset=dataset_name,
            partitioning_version="firstpass",
            root_data_dir=data_dir,
            full_path=True,
            sharded=True,
        )
    val_paths = []
    for partition_name in val_partition.split("+"):
        if len(partition_name) == 0:
            continue
        val_paths += get_partition_list(
            partition_name,
            dataset=dataset_name,
            partitioning_version="firstpass",
            root_data_dir=data_dir,
            full_path=True,
            sharded=True,
        )
    print(
        "Found {:3d} train sample paths from partition {} for dataset {}".format(
            len(train_paths), train_partition, dataset_name
        )
    )
    print(
        "Found {:3d} val sample paths from partition {} for dataset {}".format(
            len(val_paths), val_partition, dataset_name
        )
    )

    dataset_args = {}
    if dataset_name == "mobile":
        dataset_args["remove_nearfield"] = False
        dataset_args["remove_offset_turbulence"] = 0
        dataset_args["remove_offset_bottom"] = 1.0
    elif dataset_name == "MinasPassage":
        dataset_args["remove_nearfield"] = True
        # NB: Nearfield distance is 1.7, except for the samples:
        # march2018_D20180502-T045216_D20180502-T102215
        # march2018_D20180502-T105216_D20180502-T162215
        # march2018_D20180502-T165215_D20180502-T222216
        # march2018_D20180502-T225214_D20180503-T042215
        # For which it is 1.744
        dataset_args["nearfield_distance"] = 1.745
        dataset_args["remove_offset_turbulence"] = 0
        dataset_args["remove_offset_bottom"] = 0
    elif dataset_name == "GrandPassage":
        dataset_args["remove_nearfield"] = True
        dataset_args["nearfield_distance"] = 1.7
        dataset_args["remove_offset_turbulence"] = 1.0
        dataset_args["remove_offset_bottom"] = 0

    dataset_train = echofilter.data.dataset.TransectDataset(
        train_paths,
        window_len=sample_shape[0],
        p_scale_window=0.8,
        num_windows_per_transect=None,
        use_dynamic_offsets=True,
        crop_depth=crop_depth,
        transform=train_transform,
        **dataset_args,
    )
    dataset_val = echofilter.data.dataset.TransectDataset(
        val_paths,
        window_len=sample_shape[0],
        p_scale_window=0,
        num_windows_per_transect=None,
        use_dynamic_offsets=False,
        crop_depth=crop_depth,
        transform=val_transform,
        **dataset_args,
    )
    dataset_augval = echofilter.data.dataset.TransectDataset(
        val_paths,
        window_len=sample_shape[0],
        p_scale_window=0.8,
        num_windows_per_transect=None,
        use_dynamic_offsets=False,
        crop_depth=crop_depth,
        transform=train_transform,
        **dataset_args,
    )
    return dataset_train, dataset_val, dataset_augval


def train_epoch(
    loader,
    model,
    criterion,
    optimizer,
    device,
    epoch,
    dtype=torch.float,
    print_freq=10,
    schedule_data=None,
    use_mixed_precision=False,
    continue_through_error=True,
):
    """
    Train a model through a single epoch of the dataset.

    Parameters
    ----------
    loader : iterable, torch.utils.data.DataLoader
        Dataloader.
    model : callable, echofilter.nn.wrapper.Echofilter
        Model.
    criterion : callable, torch.nn.modules.loss._Loss
        Loss function.
    device : str or torch.device
        Which device the data should be loaded onto.
    epoch : int
        Which epoch is being performed.
    dtype : str or torch.dtype
        Datatype which which the data should be loaded.
    print_freq : int, optional
        Number of batches between reporting progress. Default is ``10``.
    schedule_data : dict or None
        If a learning rate schedule is being used, this may be passed as a
        dictionary with the key ``"scheduler"`` mapping to the learning rate
        schedule as a callable.
    use_mixed_precision : bool
        Whether to use :meth:`apex.amp.scale_loss` to automatically scale the
        loss. Default is ``False``.
    continue_through_error : bool
        Whether to catch errors within an individual batch, ignore them and
        continue running training on the rest of the batches. If there are
        five or more errors while processing the batch, training will halt
        regardless of ``continue_through_error``. Default is ``True``.

    Returns
    -------
    average_loss : float
        Average loss as given by criterion (weighted equally for each sample
        in ``loader``).
    meters : dict of dict
        Each key is a strata of the model output, each mapping to a their own
        dictionary of evaluation criterions: "Accuracy", "Precision", "Recall",
        "F1 Score", "Jaccard".
    examples : tuple of torch.Tensor
        Tuple of `(example_input, example_data, example_output)`.
    timing : tuple of floats
        Tuple of `(batch_time, data_time)`.
    """
    if schedule_data is None:
        schedule_data = {"name": "constant"}

    batch_time = AverageMeter("Time", ":6.3f")
    data_time = AverageMeter("Data", ":6.3f")
    losses = AverageMeter("Loss", ":6.3f")

    meters = {}
    for chn in [
        "Overall",
        "Turbulence",
        "Bottom",
        "RemovedSeg",
        "Passive",
        "Patch",
        "Surface",
    ]:
        for condition in model.conditions:
            cs = condition
            if condition != "":
                cs = "|" + condition
            cc = chn + cs
            meters[cc] = {}
            meters[cc]["Accuracy"] = AverageMeter("Accuracy (" + cc + ")", ":6.2f")
            meters[cc]["Precision"] = AverageMeter("Precision (" + cc + ")", ":6.2f")
            meters[cc]["Recall"] = AverageMeter("Recall (" + cc + ")", ":6.2f")
            meters[cc]["F1 Score"] = AverageMeter("F1 Score (" + cc + ")", ":6.4f")
            meters[cc]["Jaccard"] = AverageMeter("Jaccard (" + cc + ")", ":6.4f")

    progress = ProgressMeter(
        len(loader),
        [
            batch_time,
            data_time,
            losses,
            meters["Overall"]["Accuracy"],
            meters["Overall"]["Jaccard"],
        ],
        prefix="Epoch: [{}]".format(epoch),
    )

    # switch to train mode
    model.train()

    example_input = example_data = example_output = None
    n_backward_errors = 0

    end = time.time()
    for i, (input, metadata) in enumerate(loader):
        # measure data loading time
        data_time.update(time.time() - end)

        input = input.to(device, dtype, non_blocking=True)
        metadata = {
            k: v.to(device, dtype, non_blocking=True) for k, v in metadata.items()
        }

        # Compute output
        output = model(input)
        loss = criterion(output, metadata)
        # Record loss
        ns = input.size(0)
        losses.update(loss.item(), ns)

        with torch.no_grad():
            if i == max(0, len(loader) - 2):
                example_input = input.detach()
                example_data = {k: v.detach() for k, v in metadata.items()}
                example_output = output.detach()

            # Measure and record performance with various metrics
            for chn_cond, meters_k in meters.items():

                chnparts = chn_cond.split("|")
                chn = chnparts[0].lower()
                if len(chnparts) < 2:
                    cs = cond = ""
                else:
                    cond = chnparts[1]
                    cs = "|" + cond

                if chn.startswith("overall"):
                    output_k = output["mask_keep_pixel" + cs].float()
                    target_k = metadata["mask"]
                elif chn.startswith("turbulence"):
                    output_k = output["p_is_below_turbulence" + cs]
                    target_k = 1 - metadata["mask_turbulence"]
                elif chn.startswith("surf"):
                    output_k = output["p_is_below_surface" + cs]
                    target_k = 1 - metadata["mask_surface"]
                elif chn.startswith("bottom"):
                    output_k = output["p_is_above_bottom" + cs]
                    target_k = 1 - metadata["mask_bottom"]
                elif chn.startswith("removedseg"):
                    output_k = output["p_is_removed" + cs]
                    target_k = metadata["is_removed"]
                elif chn.startswith("passive"):
                    output_k = output["p_is_passive" + cs]
                    target_k = metadata["is_passive"]
                elif chn.startswith("patch"):
                    output_k = output["p_is_patch" + cs]
                    target_k = metadata["mask_patches"]
                else:
                    raise ValueError("Unrecognised output channel: {}".format(chn))

                if cond:
                    if cond.startswith("up"):
                        mask = metadata["is_upward_facing"] > 0.5
                    elif cond.startswith("down"):
                        mask = metadata["is_upward_facing"] < 0.5
                    else:
                        raise ValueError("Unsupported condition {}".format(cond))
                    if torch.sum(mask).item() == 0:
                        continue
                    output_k = output_k[mask]
                    target_k = target_k[mask]

                for c, v in meters_k.items():
                    c = c.lower()
                    if c == "accuracy":
                        v.update(
                            100.0 * criterions.mask_accuracy(output_k, target_k).item(),
                            ns,
                        )
                    elif c == "precision":
                        v.update(
                            100.0
                            * criterions.mask_precision(output_k, target_k).item(),
                            ns,
                        )
                    elif c == "recall":
                        v.update(
                            100.0 * criterions.mask_recall(output_k, target_k).item(),
                            ns,
                        )
                    elif c == "f1 score" or c == "f1":
                        v.update(
                            criterions.mask_f1_score(output_k, target_k).item(), ns
                        )
                    elif c == "jaccard":
                        v.update(
                            criterions.mask_jaccard_index(output_k, target_k).item(), ns
                        )
                    elif c == "active output":
                        v.update(
                            100.0 * criterions.mask_active_fraction(output_k).item(), ns
                        )
                    elif c == "active target":
                        v.update(
                            100.0 * criterions.mask_active_fraction(target_k).item(), ns
                        )
                    else:
                        raise ValueError("Unrecognised criterion: {}".format(c))

        # compute gradient and do optimizer update step
        optimizer.zero_grad()

        try:
            if use_mixed_precision:
                with apex.amp.scale_loss(loss, optimizer) as scaled_loss:
                    scaled_loss.backward()
            else:
                loss.backward()
        except Exception as ex:
            n_backward_errors += 1
            if n_backward_errors > 5:
                # If there have been more than 5 faulty batches in this epoch, we halt
                # now since it seems like a systemic problem and not a one-off.
                continue_through_error = False
            if not continue_through_error:
                raise ex
            print("Error in backward step:")
            print("".join(traceback.TracebackException.from_exception(ex).format()))

        optimizer.step()
        if "scheduler" in schedule_data:
            schedule_data["scheduler"].step()

        # measure elapsed time
        batch_time.update(time.time() - end)
        end = time.time()

        if i % print_freq == 0 or i + 1 == len(loader):
            progress.display(i + 1)

    return (
        losses.avg,
        meters,
        (example_input, example_data, example_output),
        (batch_time, data_time),
    )


def validate(
    loader,
    model,
    criterion,
    device,
    dtype=torch.float,
    print_freq=10,
    prefix="Test",
    num_examples=32,
):
    """
    Validate the model's performance on the validation partition.

    Parameters
    ----------
    loader : iterable, torch.utils.data.DataLoader
        Dataloader.
    model : callable, echofilter.nn.wrapper.Echofilter
        Model.
    criterion : callable, torch.nn.modules.loss._Loss
        Loss function.
    device : str or torch.device
        Which device the data should be loaded onto.
    dtype : str or torch.dtype
        Datatype which which the data should be loaded.
    print_freq : int, optional
        Number of batches between reporting progress. Default is ``10``.
    prefix : str, optional
        Prefix string to prepend to progress meter names. Default is ``"Test"``.
    num_examples : int, optional
        Number of example inputs to return. Default is ``32``.

    Returns
    -------
    average_loss : float
        Average loss as given by criterion (weighted equally for each sample
        in ``loader``).
    meters : dict of dict
        Each key is a strata of the model output, each mapping to a their own
        dictionary of evaluation criterions: "Accuracy", "Precision", "Recall",
        "F1 Score", "Jaccard".
    examples : tuple of torch.Tensor
        Tuple of `(example_input, example_data, example_output)`.
    """
    batch_time = AverageMeter("Time", ":6.3f")
    data_time = AverageMeter("Data", ":6.3f")
    losses = AverageMeter("Loss", ":6.3f")

    meters = {}
    for chn in [
        "Overall",
        "Turbulence",
        "Bottom",
        "RemovedSeg",
        "Passive",
        "Patch",
        "Surface",
    ]:
        for condition in model.conditions:
            cs = condition
            if condition != "":
                cs = "|" + condition
            cc = chn + cs
            meters[cc] = {}
            meters[cc]["Accuracy"] = AverageMeter("Accuracy (" + cc + ")", ":6.2f")
            meters[cc]["Precision"] = AverageMeter("Precision (" + cc + ")", ":6.2f")
            meters[cc]["Recall"] = AverageMeter("Recall (" + cc + ")", ":6.2f")
            meters[cc]["F1 Score"] = AverageMeter("F1 Score (" + cc + ")", ":6.4f")
            meters[cc]["Jaccard"] = AverageMeter("Jaccard (" + cc + ")", ":6.4f")

    progress = ProgressMeter(
        len(loader),
        [
            batch_time,
            data_time,
            losses,
            meters["Overall"]["Accuracy"],
            meters["Overall"]["Jaccard"],
        ],
        prefix=prefix + ": ",
    )

    # switch to evaluate mode
    model.eval()

    example_input = []
    example_data = []
    example_output = []
    example_interval = max(1, len(loader) // num_examples)

    with torch.no_grad():
        end = time.time()
        for i, (input, metadata) in enumerate(loader):
            # measure data loading time
            data_time.update(time.time() - end)

            input = input.to(device, dtype, non_blocking=True)
            metadata = {
                k: v.to(device, dtype, non_blocking=True) for k, v in metadata.items()
            }

            # Compute output
            output = model(input)
            loss = criterion(output, metadata)
            # Record loss
            ns = input.size(0)
            losses.update(loss.item(), ns)

            if i % example_interval == 0 and len(example_input) < num_examples:
                example_input.append(input[0].detach())
                example_data.append({k: v[0].detach() for k, v in metadata.items()})
                example_output.append({k: v[0].detach() for k, v in output.items()})

            # Measure and record performance with various metrics
            for chn_cond, meters_k in meters.items():

                chnparts = chn_cond.split("|")
                chn = chnparts[0].lower()
                if len(chnparts) < 2:
                    cs = cond = ""
                else:
                    cond = chnparts[1]
                    cs = "|" + cond

                if chn.startswith("overall"):
                    output_k = output["mask_keep_pixel" + cs].float()
                    target_k = metadata["mask"]
                elif chn.startswith("turbulence"):
                    output_k = output["p_is_below_turbulence" + cs]
                    target_k = 1 - metadata["mask_turbulence"]
                elif chn.startswith("surf"):
                    output_k = output["p_is_below_surface" + cs]
                    target_k = 1 - metadata["mask_surface"]
                elif chn.startswith("bottom"):
                    output_k = output["p_is_above_bottom" + cs]
                    target_k = 1 - metadata["mask_bottom"]
                elif chn.startswith("removedseg"):
                    output_k = output["p_is_removed" + cs]
                    target_k = metadata["is_removed"]
                elif chn.startswith("passive"):
                    output_k = output["p_is_passive" + cs]
                    target_k = metadata["is_passive"]
                elif chn.startswith("patch"):
                    output_k = output["p_is_patch" + cs]
                    target_k = metadata["mask_patches"]
                else:
                    raise ValueError("Unrecognised output channel: {}".format(chn))

                if cond:
                    if cond.startswith("up"):
                        mask = metadata["is_upward_facing"] > 0.5
                    elif cond.startswith("down"):
                        mask = metadata["is_upward_facing"] < 0.5
                    else:
                        raise ValueError("Unsupported condition {}".format(cond))
                    if torch.sum(mask).item() == 0:
                        continue
                    output_k = output_k[mask]
                    target_k = target_k[mask]

                for c, v in meters_k.items():
                    c = c.lower()
                    if c == "accuracy":
                        v.update(
                            100.0 * criterions.mask_accuracy(output_k, target_k).item(),
                            ns,
                        )
                    elif c == "precision":
                        v.update(
                            100.0
                            * criterions.mask_precision(output_k, target_k).item(),
                            ns,
                        )
                    elif c == "recall":
                        v.update(
                            100.0 * criterions.mask_recall(output_k, target_k).item(),
                            ns,
                        )
                    elif c == "f1 score" or c == "f1":
                        v.update(
                            criterions.mask_f1_score(output_k, target_k).item(), ns
                        )
                    elif c == "jaccard":
                        v.update(
                            criterions.mask_jaccard_index(output_k, target_k).item(), ns
                        )
                    elif c == "active output":
                        v.update(
                            100.0 * criterions.mask_active_fraction(output_k).item(), ns
                        )
                    elif c == "active target":
                        v.update(
                            100.0 * criterions.mask_active_fraction(target_k).item(), ns
                        )
                    else:
                        raise ValueError("Unrecognised criterion: {}".format(c))

            # measure elapsed time
            batch_time.update(time.time() - end)
            end = time.time()

            if i % print_freq == 0 or i + 1 == len(loader):
                progress.display(i + 1)

    # Restack samples, converting list into higher-dim tensor
    example_input = torch.stack(example_input, dim=0)
    example_data = {
        k: torch.stack([a[k] for a in example_data], dim=0) for k in example_data[0]
    }
    example_output = {
        k: torch.stack([a[k] for a in example_output], dim=0) for k in example_output[0]
    }

    return losses.avg, meters, (example_input, example_data, example_output)


def generate_from_transect(model, transect, sample_shape, device, dtype=torch.float):
    """
    Generate an output for a sample transect, .
    """
    # Put model in evaluation mode
    model.eval()

    # Make a copy of the transect which we will use to
    data = copy.deepcopy(transect)

    # Configure data to match what the model expects to see
    # Ensure depth is always increasing (which corresponds to descending from
    # the air down the water column)
    if data["depths"][-1] < data["depths"][0]:
        # Found some upward-facing data that still needs to be reflected
        for k in ["depths", "signals", "mask"]:
            data[k] = np.flip(data[k], -1).copy()

    # Apply transforms
    transform = torchvision.transforms.Compose(
        [
            echofilter.data.transforms.Normalize(DATA_CENTER, DATA_DEVIATION),
            echofilter.data.transforms.ReplaceNan(NAN_VALUE),
            echofilter.data.transforms.Rescale(
                (data["signals"].shape[0], sample_shape[1])
            ),
        ]
    )
    data = transform(data)
    input = torch.tensor(data["signals"]).unsqueeze(0).unsqueeze(0)
    input = input.to(device, dtype).contiguous()
    # Put data through model
    with torch.no_grad():
        output = model(input)
        output = {k: v.squeeze(0).cpu().numpy() for k, v in output.items()}

    output["depths"] = data["depths"]
    output["timestamps"] = data["timestamps"]

    return output


def _generate_from_loaded(transect, model, *args, crop_depth=None, **kwargs):
    """
    Generate an output from a loaded transect.
    """
    # Crop long input
    for key in (
        echofilter.data.transforms._fields_2d
        + echofilter.data.transforms._fields_1d_timelike
    ):
        if key in transect:
            transect[key] = transect[key][:MAX_INPUT_LEN]

    # Apply depth crop
    if crop_depth is not None:
        if transect["is_upward_facing"]:
            depth_crop_mask = (
                transect["depths"] >= np.max(transect["depths"]) - crop_depth
            )
        else:
            depth_crop_mask = transect["depths"] <= crop_depth

        for key in echofilter.data.transforms._fields_2d:
            if key in transect:
                transect[key] = transect[key][:, depth_crop_mask]

        for key in echofilter.data.transforms._fields_1d_depthlike:
            if key in transect:
                transect[key] = transect[key][depth_crop_mask]

    # Convert lines to masks
    ddepths = np.broadcast_to(transect["depths"], transect["Sv"].shape)
    transect["mask_turbulence"] = np.single(
        ddepths < np.expand_dims(transect["turbulence"], -1)
    )
    transect["mask_bottom"] = np.single(
        ddepths > np.expand_dims(transect["bottom"], -1)
    )
    # Add mask_patches to the data, for plotting
    transect["mask_patches"] = 1 - transect["mask"]
    transect["mask_patches"][transect["is_passive"] > 0.5] = 0
    transect["mask_patches"][transect["is_removed"] > 0.5] = 0
    transect["mask_patches"][transect["mask_turbulence"] > 0.5] = 0
    transect["mask_patches"][transect["mask_bottom"] > 0.5] = 0

    # Generate predictions for the transect
    transect["signals"] = transect.pop("Sv")
    prediction = generate_from_transect(model, transect, *args, **kwargs)
    transect["Sv"] = transect.pop("signals")

    return transect, prediction


def generate_from_file(fname, *args, **kwargs):
    """
    Generate an output for a sample transect, specified by its file path.
    """
    # Load the data
    transect = load_decomposed_transect_mask(fname)
    # Process the transect
    return _generate_from_loaded(transect, *args, **kwargs)


def generate_from_shards(fname, *args, **kwargs):
    """
    Generate an output for a sample transect, specified by a path to sharded data.
    """
    # Load the data
    transect = echofilter.raw.shardloader.load_transect_segments_from_shards_abs(fname)
    # Process the transect
    return _generate_from_loaded(transect, *args, **kwargs)


def save_checkpoint(state, is_best, dirname=".", fname_fmt="checkpoint{}.pt", dup=None):
    """
    Save a model checkpoint, using :meth:`torch.save`.

    Parameters
    ----------
    state : dict
        Model checkpoint state to record.
    is_best : bool
        Whether this model state is the best so far. If ``True``, the best
        checkpoint (by default named ``"checkpoint_best.pt"``) will be overwritten
        with this ``state``.
    dirname : str, optional
        Path to directory in which the checkpoint will be saved.
        Default is ``"."`` (current directory of the executed script).
    fname_fmt : str, optional
        Format for the file name(s) of the saved checkpoint(s). Must include
        one string argument output. Default is ``"checkpoint{}.pt"``.
    dup : str or None
        If this is not ``None``, a duplicate copy of the checkpoint is recorded
        in accordance with ``fname_fmt``. By default the duplicate output file
        name will be styled as ``"checkpoint_<dup>.pt"``.
    """
    os.makedirs(dirname, exist_ok=True)
    fname = os.path.join(dirname, fname_fmt.format(""))
    torch.save(state, fname)
    if is_best:
        shutil.copyfile(fname, os.path.join(dirname, fname_fmt.format("_best")))
    if dup:
        shutil.copyfile(
            fname, os.path.join(dirname, fname_fmt.format("_{}".format(dup)))
        )


def meters_to_csv(meters, is_best, dirname=".", filename="meters.csv"):
    """
    Export performance metrics to CSV format.

    Parameters
    ----------
    meters : dict of dict
        Collection of output meters, as a nested dictionary.
    is_best : bool
        Whether this model state is the best so far. If ``True``, the CSV file
        will be copied to ``"model_best.meters.csv"``.
    dirname : str, optional
        Path to directory in which the checkpoint will be saved.
        Default is ``"."`` (current directory of the executed script).
    filename : str, optional
        Format for the output file. Default is ``"meters.csv"``.
    """
    os.makedirs(dirname, exist_ok=True)
    df = pd.DataFrame()
    for chn in meters:
        if "|" in chn:
            # Skip conditional model evaluations
            continue
        # For each output plane
        for _, meter in meters[chn].items():
            # For each criterion
            df[meter.name] = meter.values
    df.to_csv(os.path.join(dirname, filename), index=False)
    if is_best:
        shutil.copyfile(
            os.path.join(dirname, filename),
            os.path.join(dirname, "model_best.meters.csv"),
        )


if __name__ == "__main__":
    main()



"""
Window management for Windows.
"""

# This file is part of Echofilter.
#
# Copyright (C) 2020-2022  Scott C. Lowe and Offshore Energy Research Association (OERA)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import re
from contextlib import contextmanager

from .. import ui

try:
    import pywintypes
    import win32com.client
    import win32con
    import win32gui
except ImportError:
    from ..path import check_if_windows

    if check_if_windows():
        raise

    import warnings

    msg = "The Windows management module is only for Windows operating systems."
    with ui.style.warning_message(msg) as msg:
        warnings.warn(msg, category=RuntimeWarning)


__all__ = ["opencom", "WindowManager"]


class WindowManager:
    """
    Encapsulates calls to window management using the Windows api.

    Notes
    -----
    Based on: https://stackoverflow.com/a/2091530 and https://stackoverflow.com/a/4440622
    """

    def __init__(self, title=None, class_name=None, title_pattern=None):
        self.handle = None
        if title is not None:
            self.find_window(class_name=class_name, title=title)
        elif title_pattern is not None:
            self.find_window_regex(title_pattern)

    def find_window(self, class_name=None, title=None):
        """Find a window by its exact title."""
        handle = win32gui.FindWindow(class_name, title)
        if handle == 0:
            raise EnvironmentError(
                "Couldn't find window with class={}, title={}.".format(
                    class_name, title
                )
            )
        else:
            self.handle = handle

    def _window_enum_callback(self, hwnd, pattern):
        """Pass to win32gui.EnumWindows() to check all the opened windows."""
        if re.match(pattern, str(win32gui.GetWindowText(hwnd))) is not None:
            self.handle = hwnd

    def find_window_regex(self, pattern):
        """Find a window whose title matches a regular expression."""
        self.handle = None
        win32gui.EnumWindows(self._window_enum_callback, pattern)
        if self.handle is None:
            raise EnvironmentError(
                "Couldn't find a window with title matching pattern {}.".format(pattern)
            )

    def set_foreground(self):
        """Bring the window to the foreground."""
        win32gui.SetForegroundWindow(self.handle)

    def hide(self):
        """Hide the window."""
        win32gui.ShowWindow(self.handle, win32con.SW_HIDE)

    def show(self):
        """Show the window."""
        win32gui.ShowWindow(self.handle, win32con.SW_SHOW)


@contextmanager
def opencom(
    com_name,
    can_make_anew=False,
    title=None,
    title_pattern=None,
    minimize=False,
    hide="never",
):
    """
    Open a connection to an application with a COM object.

    The application may or may not be open before this context begins. If it
    was not already open, the application is closed when leaving the context.

    Parameters
    ----------
    com_name : str
        Name of COM object to dispatch.
    can_make_anew : bool, optional
        Whether arbitrarily many sessions of the COM object can be created, and
        if so whether they should be. Default is ``False``, in which case the
        context manager will check to see if the application is already running
        before connecting to it. If it was already running, it will not be
        closed when this context closes.
    title : str, optional
        Exact title of window. If the title can not be determined exactly, use
        ``title_pattern`` instead.
    title_pattern : str, optional
        Regular expression for the window title.
    minimize : bool, optional
        If ``True``, the application will be minimized while the code runs.
        Default is ``False``.
    hide : {"never", "new", "always"}, optional
        Whether to hide the application window entirely. Default is ``"never"``.
        If this is enabled, at least one of ``title`` and ``title_pattern`` must
        be specified.  If ``hide="new"``, the application is only hidden if it
        was created by this context, and not if it was already running.
        If ``hide="always"``, the application is hidden even if it was already
        running. In the latter case, the window will be revealed again when
        leaving this context.

    Yields
    ------
    win32com.gen_py
        Interface to COM object.
    """
    HIDE_OPTIONS = {"never", "new", "always"}
    if hide not in HIDE_OPTIONS:
        raise ValueError(
            "Unsupported hide value: {}. Must be one of {}".format(hide, HIDE_OPTIONS)
        )

    make_anew = can_make_anew
    if not make_anew:
        # Try to fetch existing session
        try:
            app = win32com.client.GetActiveObject(com_name)
            existing_session = True
        except pywintypes.com_error:
            # No existing session, make a new session
            make_anew = True
    if make_anew:
        # Create a new session
        existing_session = False
        app = win32com.client.Dispatch(com_name)

    was_minimized = False
    if minimize:
        # Tell the app to minimize itself
        try:
            app.Minimize()
            was_minimized = True
        except Exception:
            print(ui.style.warning_fmt("Could not minimize {} window".format(com_name)))

    was_hidden = False
    if hide == "always" or (hide == "new" and not existing_session):
        # Find handle to the app based on the window title, and hide it
        winman = WindowManager()
        if title is None and title_pattern is None:
            raise ValueError(
                "One of the arguments title or title_pattern must be set in"
                " order to hide the window."
            )
        if title is not None:
            try:
                winman.find_window(title=title)
            except Exception:
                pass
        if winman.handle is None and title_pattern is not None:
            try:
                winman.find_window_regex(title_pattern)
            except Exception:
                pass
        if winman.handle is None:
            print(
                ui.style.warning_fmt(
                    "Could not hide {} window with title {}".format(
                        com_name, title if title_pattern is None else title_pattern
                    )
                )
            )
        else:
            winman.hide()
            was_hidden = True

    try:
        yield app
    finally:
        # As we leave the context, fix the state of the app to how it was
        # before we started
        command = ""
        try:
            if not existing_session:
                # If we opened it, tell the application to quit
                command = "exit"
                app.Quit()
            else:
                if was_minimized:
                    # Restore the window from being minimised
                    command = "restore"
                    app.Restore()
                if was_hidden:
                    # Show the window again
                    command = "unhide"
                    winman.show()
        except Exception:
            # We'll get an error if the application was already closed, etc
            print(
                ui.style.warning_fmt(
                    "Could not {} the {} window with handle {}.".format(
                        command, com_name, app
                    )
                )
            )



"""
Echoview interface management.
"""

# This file is part of Echofilter.
#
# Copyright (C) 2020-2022  Scott C. Lowe and Offshore Energy Research Association (OERA)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
from contextlib import contextmanager

from .. import ui

__all__ = ["ECHOVIEW_COM_NAME", "maybe_open_echoview", "open_ev_file"]

ECHOVIEW_COM_NAME = "EchoviewCom.EvApplication"


@contextmanager
def maybe_open_echoview(
    app=None,
    do_open=True,
    minimize=False,
    hide="new",
):
    """
    If the current pointer to the Echoview is invalid, open an Echoview window.

    Parameters
    ----------
    app : COM object or None, optional
        Existing COM object to interface with Echoview.
    do_open : bool, optional
        If ``False`` (dry-run mode), we don't actually need Echoview open and so
        don't try to open it. In this case, ``None`` is yielded. Present so a
        context manager can be used even if the application isn't opened.
        Default is ``True``, do open Echoview.
    minimize : bool, optional
        If ``True``, the Echoview window being used will be minimized while the
        code runs. Default is ``False``.
    hide : {"never", "new", "always"}, optional
        Whether to hide the Echoview window entirely. If ``hide="new"``, the
        application is only hidden if it was created by this context, and not
        if it was already running. If ``hide="always"``, the application is
        hidden even if it was already running. In the latter case, the window
        will be revealed again when leaving this context. Default is ``"new"``.
    """
    if not do_open:
        yield None
        return

    if app is None:
        need_to_open = True
    else:
        try:
            # If we can check for a license, the COM handle is working
            app.IsLicensed()
            need_to_open = False
        except Exception:
            need_to_open = True
    if not need_to_open:
        yield app
    else:
        # Need to import the context which actually opens COM windows
        from .manager import opencom

        with opencom(
            ECHOVIEW_COM_NAME,
            title="Echoview",
            title_pattern="Echoview.*",
            minimize=minimize,
            hide=hide,
        ) as app:
            yield app


@contextmanager
def open_ev_file(filename, app=None):
    """
    Open an EV file within a context.

    Parameters
    ----------
    filename : str
        Path to file to open.
    app : COM object or None, optional
        Existing COM object to interface with Echoview. If ``None``, a new
        COM interface is created. If that requires opening a new instance
        of Echoview, it is hidden while the file is in use.
    """
    if not os.path.isfile(filename):
        raise EnvironmentError("Missing file: {}".format(filename))
    # Ensure the path is an absolute path
    filename = os.path.abspath(filename)
    with maybe_open_echoview(app, hide="new") as app:
        ev_file = app.OpenFile(filename)
        if ev_file is None or (
            isinstance(ev_file, str) and ev_file in ("None", "NoneValue")
        ):
            raise EnvironmentError(
                "Could not open file '{}'."
                "\nIt appears that you already have a file open in Echoview."
                " {}Please close Echoview before running echofilter{}."
                " You can re-open Echoview once echofilter is running and"
                " continue your work without distrupting echofilter.".format(
                    filename,
                    ui.style.HighlightStyle.start,
                    ui.style.HighlightStyle.reset,
                )
            )
        try:
            yield ev_file
        finally:
            try:
                ev_file.Close()
            except Exception:
                print(
                    ui.style.warning_fmt(
                        "Could not close Echoview file {}".format(filename)
                    )
                )



#!/usr/bin/env python
"""
Export raw EV files in CSV format.
"""

# ev2csv is part of Echofilter.
#
# Copyright (C) 2020-2022  Scott C. Lowe and Offshore Energy Research Association (OERA)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import sys
import warnings

from tqdm.auto import tqdm

import echofilter.path
import echofilter.ui
import echofilter.utils
import echofilter.win

# Provide a warning for non-Windows users
if not echofilter.path.check_if_windows():
    msg = (
        "\nev2csv requires the Echoview application, which is only"
        " available on Windows operating systems."
    )
    with echofilter.ui.style.warning_message(msg) as msg:
        print("")
        warnings.warn(msg, category=RuntimeWarning)


DEFAULT_VARNAME = "Fileset1: Sv pings T1"


def run_ev2csv(
    paths,
    variable_name=DEFAULT_VARNAME,
    export_raw=True,
    source_dir=".",
    recursive_dir_search=True,
    output_dir="",
    suffix=None,
    keep_ext=False,
    skip_existing=False,
    overwrite_existing=False,
    minimize_echoview=False,
    hide_echoview="new",
    verbose=1,
    dry_run=False,
):
    """
    Export EV files to raw CSV files.

    Parameters
    ----------
    paths : iterable
        Paths to input EV files to process, or directories containing EV files.
        These may be full paths or paths relative to ``source_dir``. For each
        folder specified, any files with extension ``"csv"`` within the folder
        and all its tree of subdirectories will be processed.
    variable_name : str, optional
        Name of the Echoview acoustic variable to export. Default is
        `"Fileset1: Sv pings T1"`.
    export_raw : bool, optional
        If ``True`` (default), exclusion and threshold settings in the EV file
        are temporarily disabled before exporting the CSV, in order to ensure
        all raw data is exported. If ``False``, thresholds and exclusions are
        used as per the EV file.
    source_dir : str, optional
        Path to directory where files are found. Default is ``"."``.
    recursive_dir_search : bool, optional
        How to handle directory inputs in ``paths``. If ``False``, only files
        (with the correct extension) in the directory will be included.
        If ``True``, subdirectories will also be walked through to find input
        files. Default is ``True``.
    output_dir : str, optional
        Directory where output files will be written. If this is an empty
        string (``""``, default), outputs are written to the same directory as
        each input file. Otherwise, they are written to ``output_dir``,
        preserving their path relative to ``source_dir`` if relative paths were
        used.
    suffix : str, optional
        Output filename suffix. Default is ``"_Sv_raw.csv"`` if ``keep_ext=False``,
        or ``".Sv_raw.csv"`` if ``keep_ext=True``. The ``"_raw"`` component is
        excluded if ``export_raw`` is ``False``.
    keep_ext : bool, optional
        Whether to preserve the file extension in the input file name when
        generating output file name. Default is ``False``, removing the
        extension.
    skip_existing : bool, optional
        Whether to skip processing files whose destination paths already
        exist. If ``False`` (default), an error is raised if the destination file
        already exists.
    overwrite_existing : bool, optional
        Whether to overwrite existing output files. If ``False`` (default), an
        error is raised if the destination file already exists.
    minimize_echoview : bool, optional
        If ``True``, the Echoview window being used will be minimized while this
        function is running. Default is ``False``.
    hide_echoview : {"never", "new", "always"}, optional
        Whether to hide the Echoview window entirely while the code runs.
        If ``hide_echoview="new"``, the application is only hidden if it
        was created by this function, and not if it was already running.
        If ``hide_echoview="always"``, the application is hidden even if it was
        already running. In the latter case, the window will be revealed again
        when this function is completed. Default is ``"new"``.
    verbose : int, optional
        Level of verbosity. Default is ``1``.
    dry_run : bool, optional
        If ``True``, perform a trial run with no changes made. Default is
        ``False``.

    Returns
    -------
    list of str
        Paths to generated CSV files.
    """
    if suffix is not None:
        pass
    elif keep_ext:
        suffix = ".Sv{}.csv".format("_raw" if export_raw else "")
    else:
        suffix = "_Sv{}.csv".format("_raw" if export_raw else "")

    files = list(
        echofilter.path.parse_files_in_folders(
            paths, source_dir, "ev", recursive=recursive_dir_search
        )
    )
    if verbose >= 1:
        print("Processing {} file{}".format(len(files), "" if len(files) == 1 else "s"))

    disable_tqdm = len(files) == 1 or verbose <= 0
    skip_count = 0
    output_files = []

    # Open Echoview connection
    with echofilter.win.maybe_open_echoview(
        do_open=not dry_run,
        minimize=minimize_echoview,
        hide=hide_echoview,
    ) as ev_app:
        for fname in tqdm(files, desc="ev2csv", disable=disable_tqdm):
            if verbose >= 2:
                print("Exporting {} to raw CSV".format(fname))

            # Check what the full path should be
            fname_full = echofilter.path.determine_file_path(fname, source_dir)

            # Determine where destination should be placed
            destination = echofilter.path.determine_destination(
                fname, fname_full, source_dir, output_dir
            )
            if not keep_ext:
                destination = os.path.splitext(destination)[0]
            destination += suffix

            # Check whether to skip processing this file
            if not os.path.exists(destination):
                pass
            elif skip_existing:
                if verbose >= 2:
                    print("Skipping {}".format(fname))
                skip_count += 1
                continue
            elif not overwrite_existing:
                raise EnvironmentError(
                    "Output {} already exists.\n"
                    " Run with overwrite_existing=True (with the command line"
                    " interface, use the --force flag) to overwrite existing"
                    " outputs, or skip_existing=True (with the command line"
                    " interface, use the --skip-existing flag) to skip existing"
                    " outputs.".format(destination)
                )

            if dry_run:
                if verbose >= 1:
                    print("Would write to CSV file to {}".format(destination))
                continue

            # Export a single EV file to raw CSV
            ev2csv(
                fname_full,
                destination,
                variable_name=variable_name,
                export_raw=export_raw,
                ev_app=ev_app,
                verbose=verbose - 1,
            )
            output_files.append(destination)

    if verbose >= 1:
        s = "Finished {}processing {} file{}.".format(
            "simulating " if dry_run else "",
            len(files),
            "" if len(files) == 1 else "s",
        )
        if skip_count > 0:
            s += " Of these, {} file{} skipped.".format(
                skip_count,
                " was" if skip_count == 1 else "s were",
            )
        print(s)

    return output_files


def ev2csv(
    input,
    destination,
    variable_name=DEFAULT_VARNAME,
    export_raw=True,
    ev_app=None,
    verbose=0,
):
    """
    Export a single EV file to CSV.

    Parameters
    ----------
    input : str
        Path to input file.
    destination : str
        Filename of output destination.
    variable_name : str, optional
        Name of the Echoview acoustic variable to export. Default is
        `"Fileset1: Sv pings T1"`.
    export_raw : bool, optional
        If ``True`` (default), exclusion and threshold settings in the EV file
        are temporarily disabled before exporting the CSV, in order to ensure
        all raw data is exported.
    ev_app : win32com.client.Dispatch object or None, optional
        An object which can be used to interface with the Echoview application,
        as returned by :class:`win32com.client.Dispatch`. If ``None`` (default), a
        new instance of the application is opened (and closed on completion).
    verbose : int, optional
        Level of verbosity. Default is ``0``.

    Returns
    -------
    destination : str
        Absolute path to ``destination``.
    """
    if verbose >= 1:
        print("  Opening {} in Echoview".format(input))

    # Ensure input and destination are absolute paths
    input = os.path.abspath(input)
    destination = os.path.abspath(destination)

    # Open the EV file
    with echofilter.win.open_ev_file(input, ev_app) as ev_file:

        # Find the right variable
        av = ev_file.Variables.FindByName(variable_name).AsVariableAcoustic()

        if export_raw:
            # Make sure we don't exclude anything, i.e. export "raw" data
            av.Properties.Analysis.ExcludeAbove = "None"
            av.Properties.Analysis.ExcludeBelow = "None"
            av.Properties.Analysis.ExcludeBadDataRegions = False
            av.Properties.Analysis.ExcludeBadLineStatusPings = False
            av.Properties.Data.ApplyMinimumThreshold = False
            av.Properties.Data.ApplyMaximumThreshold = False
            av.Properties.Data.ApplyMinimumTsThreshold = False
            av.Properties.Data.ApplyTimeVariedThreshold = False

        # Export the raw file
        if verbose >= 1:
            print("  Writing output {}".format(destination))
        os.makedirs(os.path.dirname(destination), exist_ok=True)
        av.ExportData(destination, -1, -1)

    # The file is automatically closed when we leave the context
    return destination


def get_parser():
    """
    Build parser for ev2csv command line interface.

    Returns
    -------
    parser : argparse.ArgumentParser
        CLI argument parser for ev2csv.
    """
    import argparse

    prog = os.path.split(sys.argv[0])[1]
    if prog == "__main__.py" or prog == "__main__":
        prog = os.path.split(__file__)[1]
    parser = argparse.ArgumentParser(
        prog=prog,
        description="Echoview to raw CSV exporter",
        formatter_class=echofilter.ui.formatters.FlexibleHelpFormatter,
        add_help=False,
    )

    # Actions
    group_action = parser.add_argument_group(
        "Actions",
        "These arguments specify special actions to perform. The main action"
        " of this program is supressed if any of these are given.",
    )
    group_action.add_argument(
        "-h",
        "--help",
        action="help",
        help="Show this help message and exit.",
    )
    group_action.add_argument(
        "--version",
        "-V",
        action="version",
        version="%(prog)s {version}".format(version=echofilter.__version__),
        help="Show program's version number and exit.",
    )

    # Input files
    group_positional = parser.add_argument_group("Positional arguments")
    group_positional.add_argument(
        "paths",
        type=str,
        nargs="+",
        default=[],
        metavar="FILE_OR_DIRECTORY",
        help="""d|
            File(s)/directory(ies) to process.
            Inputs can be absolute paths or relative paths to
            either files or directories. Paths can be given
            relative to the current directory, or optionally be
            relative to the SOURCE_DIR argument specified with
            ``--source-dir``. For each directory given, the directory
            will be searched recursively for files bearing an
            extension specified by SEARCH_EXTENSION (see the
            ``--extension`` argument for details).
            Multiple files and directories can be specified,
            separated by spaces.
            This is a required argument. At least one input file
            or directory must be given.
            In order to process the directory given by SOURCE_DIR,
            specify "." for this argument, such as::

                ev2csv . --source-dir SOURCE_DIR
        """,
    )
    group_infile = parser.add_argument_group(
        "Input file arguments",
        "Optional parameters specifying which files will processed.",
    )
    group_infile.add_argument(
        "--source-dir",
        "-d",
        dest="source_dir",
        type=str,
        default=".",
        metavar="SOURCE_DIR",
        help="""
            Path to source directory which contains the files and folders
            specified by the paths argument. Default: "%(default)s" (the
            current directory).
        """,
    )
    group_infile.add_argument(
        "--recursive-dir-search",
        dest="recursive_dir_search",
        action="store_true",
        default=True,
        help="""d|
            For any directories provided in the FILE_OR_DIRECTORY
            input, all subdirectories will also be recursively
            walked through to find files to process.
            This is the default behaviour.
        """,
    )
    group_infile.add_argument(
        "--no-recursive-dir-search",
        dest="recursive_dir_search",
        action="store_false",
        help="""
            For any directories provided in the FILE_OR_DIRECTORY
            input, only files within the specified directory will
            be included in the files to process. Subfolders within
            the directory will not be included.
        """,
    )
    group_infile.add_argument(
        "--skip-existing",
        "--skip",
        dest="skip_existing",
        action="store_true",
        help="""
            Skip processing files for which all outputs already exist
        """,
    )

    # Processing arguments
    group_inproc = parser.add_argument_group(
        "Processing arguments",
        "Optional parameters specifying how to process files.",
    )
    group_inproc.add_argument(
        "--keep-exclusions",
        "--keep-thresholds",
        dest="export_raw",
        action="store_false",
        help="""
            Export CSV with all thresholds, exclusion regions, and bad data
            exclusions set as per the EV file. Default behavior is to
            ignore these settings and export the underlying raw data.
        """,
    )

    # Output files
    group_outfile = parser.add_argument_group(
        "Destination file arguments",
        "Optional parameters specifying where output files will be located.",
    )
    group_outfile.add_argument(
        "--output-dir",
        "-o",
        metavar="OUTPUT_DIR",
        type=str,
        default="",
        help="""
            Path to output directory. If empty (default), each output is placed
            in the same directory as its input file. If OUTPUT_DIR is
            specified, the full output path for each file all contains the
            subtree of the input file relative to the base directory given by
            SOURCE_DIR.
        """,
    )
    group_outfile.add_argument(
        "--dry-run",
        "-n",
        action="store_true",
        help="""
            Perform a trial run, with no changes made. Text printed to the
            command prompt indicates which files would be processed, but work
            is only simulated and not performed.
        """,
    )
    group_outfile.add_argument(
        "--force",
        "-f",
        dest="overwrite_existing",
        action="store_true",
        help="""
            Overwrite existing files without warning. Default behaviour is to
            stop processing if an output file already exists.
        """,
    )
    group_outfile.add_argument(
        "--keep-ext",
        action="store_true",
        help="""
            If provided, the output file names (evl, evr, csv) maintain the
            input file extension before their suffix (including a new file
            extension). Default behaviour is to strip the input file name
            extension before constructing the output paths.
        """,
    )
    group_outfile.add_argument(
        "--output-suffix",
        "--suffix",
        dest="suffix",
        type=str,
        default=None,
        help="""
            Output filename suffix. Default is ``"_Sv_raw.csv"``, or
            ``".Sv_raw.csv"`` if the ``--keep_ext`` argument is supplied.
            if ``--keep-exclusions`` is given, the ``"_raw"`` component is
            dropped.
        """,
    )

    # Input data transforms
    group_inproc = parser.add_argument_group(
        "Input processing arguments",
        "Optional parameters specifying how data will be loaded from the input"
        " files and transformed before it given to the model.",
    )
    group_inproc.add_argument(
        "--variable-name",
        "--vn",
        dest="variable_name",
        type=str,
        default=DEFAULT_VARNAME,
        help="""
            Name of the Echoview acoustic variable to load from EV files.
            Default: "%(default)s".
        """,
    )

    # Echoview interaction arguments
    group_evwin = parser.add_argument_group(
        "Echoview window management",
        "Optional parameters specifying how to interact with any Echoview"
        " windows which are used during this process.",
    )
    group_evwin_hiding = group_evwin.add_mutually_exclusive_group()
    group_evwin_hiding.add_argument(
        "--hide-echoview",
        dest="hide_echoview",
        action="store_const",
        const="new",
        help="""
            Hide any Echoview window spawned by this program. If it must use
            an Echoview instance which was already running, that window is not
            hidden. This is the default behaviour.
        """,
    )
    group_evwin_hiding.add_argument(
        "--show-echoview",
        dest="hide_echoview",
        action="store_const",
        const="never",
        default=None,
        help="""
            Don't hide an Echoview window created to run this code. (Disables
            the default behaviour which is equivalent to ``--hide-echoview``.)
        """,
    )
    group_evwin_hiding.add_argument(
        "--always-hide-echoview",
        "--always-hide",
        dest="hide_echoview",
        action="store_const",
        const="always",
        help="""
            Hide the Echoview window while this code runs, even if this
            process is utilising an Echoview window which was already open.
        """,
    )
    group_evwin.add_argument(
        "--minimize-echoview",
        dest="minimize_echoview",
        action="store_true",
        help="""
            Minimize any Echoview window used to runs this code while it runs.
            The window will be restored once the program is finished.
            If this argument is supplied, ``--show-echoview`` is implied unless
            ``--hide-echoview`` is also given.
        """,
    )

    # Verbosity controls
    group_verb = parser.add_argument_group(
        "Verbosity arguments",
        "Optional parameters controlling how verbose the program should be"
        " while it is running.",
    )
    group_verb.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=1,
        help="""
            Increase the level of verbosity of the program. This can be
            specified multiple times, each will increase the amount of detail
            printed to the terminal. The default verbosity level is %(default)s.
        """,
    )
    group_verb.add_argument(
        "--quiet",
        "-q",
        action="count",
        default=0,
        help="""
            Decrease the level of verbosity of the program. This can be
            specified multiple times, each will reduce the amount of detail
            printed to the terminal.
        """,
    )

    return parser


def _get_parser_sphinx():
    """
    Pre-format parser help for sphinx-argparse processing.
    """
    return echofilter.ui.formatters.format_parser_for_sphinx(get_parser())


def main(args=None):
    """
    Run ev2csv command line interface.
    """
    parser = get_parser()
    kwargs = vars(parser.parse_args(args))
    kwargs["verbose"] -= kwargs.pop("quiet", 0)

    if kwargs["hide_echoview"] is None:
        kwargs["hide_echoview"] = "never" if kwargs["minimize_echoview"] else "new"

    run_ev2csv(**kwargs)


if __name__ == "__main__":
    main()



"""
Evaluation criterions.
"""

# This file is part of Echofilter.
#
# Copyright (C) 2020-2022  Scott C. Lowe and Offshore Energy Research Association (OERA)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import torch


def _binarise_and_reshape(arg, threshold=0.5, ndim=None):
    """
    Binarize and partially flatten a tensor.

    Parameters
    ----------
    arg : array_like
        Input tensor or array.
    threshold : float, optional
        Threshold which entries in ``arg`` must exceed. Default is ``0.5``.
    ndim : int or None
        Number of dimensions to keep. If ``None``, only the first (batch)
        dimension is kept and the rest are flattened. Default is ``None``.

    Returns
    -------
    array_like
        A :class:`numpy.ndarray` or :class:`torch.Tensor` (corresponding to
        the type of ``arg``), but partially flattened and binarised.
    """
    # Binarise mask
    arg = arg > threshold
    # Reshape so pixels are vectorised by batch
    if ndim is None:
        shape = [arg.shape[0], -1]
    else:
        shape = list(arg.shape)
        shape = shape[:-ndim] + [-1]
    arg = arg.reshape(shape)
    return arg


def mask_active_fraction(input, threshold=0.5, ndim=None, reduction="mean"):
    """
    Measure the fraction of input which exceeds a threshold.

    Parameters
    ----------
    input : torch.Tensor
        Input tensor.
    threshold : float, optional
        Threshold which entries in ``input`` must exceed. Default is ``0.5``.
    ndim : int or None
        Number of dimensions to keep. If ``None``, only the first (batch)
        dimension is kept and the rest are flattened. Default is ``None``.
    reduction : ``"none"`` or ``"mean"`` or ``"sum"``, optional
        Specifies the reduction to apply to the output:
        ``"none"`` | ``"mean"`` | ``"sum"``.
        ``"none"``: no reduction will be applied,
        ``"mean"``: the sum of the output will be divided by the number of
        elements in the output,
        ``"sum"``: the output will be summed.
        Default: ``"mean"``.

    Returns
    -------
    torch.Tensor
        The fraction of ``input`` which exceeds ``threshold``, with shaped
        corresponding to ``reduction``.
    """
    # Binarise and reshape mask
    input = _binarise_and_reshape(input, threshold=threshold, ndim=ndim)

    # Measure hit rate and number of samples
    output = input.sum(-1).float() / input.size(-1)

    # Apply reduction
    if reduction == "none":
        return output
    elif reduction == "mean":
        return output.mean()
    elif reduction == "sum":
        return output.sum()
    else:
        raise ValueError("Unsupported reduction value: {}".format(reduction))


def mask_active_fraction_with_logits(input, *args, **kwargs):
    """
    Convert logits to probabilities, and measure what fraction exceed threshold.

    See Also
    --------
    mask_active_fraction
    """
    return mask_active_fraction(torch.sigmoid(input), *args, **kwargs)


def mask_accuracy(input, target, threshold=0.5, ndim=None, reduction="mean"):
    """
    Measure accuracy of input compared to binary targets.

    Parameters
    ----------
    input : torch.Tensor
        Input tensor.
    target : torch.Tensor
        Target tensor, the same shape as ``input``.
    threshold : float, optional
        Threshold which entries in ``input`` and ``target`` must exceed to be
        binarised as the positive class. Default is ``0.5``.
    ndim : int or None
        Number of dimensions to keep. If ``None``, only the first (batch)
        dimension is kept and the rest are flattened. Default is ``None``.
    reduction : ``"none"`` or ``"mean"`` or ``"sum"``, optional
        Specifies the reduction to apply to the output:
        ``"none"`` | ``"mean"`` | ``"sum"``.
        ``"none"``: no reduction will be applied,
        ``"mean"``: the sum of the output will be divided by the number of
        elements in the output,
        ``"sum"``: the output will be summed.
        Default: ``"mean"``.

    Returns
    -------
    torch.Tensor
        The fraction of ``input`` which has the same class as ``target`` after
        thresholding.
    """
    # Binarise and reshape masks
    input = _binarise_and_reshape(input, threshold=threshold, ndim=ndim)
    target = _binarise_and_reshape(target, threshold=threshold, ndim=ndim)

    # Measure hit rate and number of samples
    hits = (input == target).sum(-1)
    count = input.size(-1)
    output = hits.float() / count

    # Apply reduction
    if reduction == "none":
        return output
    elif reduction == "mean":
        return output.mean()
    elif reduction == "sum":
        return output.sum()
    else:
        raise ValueError("Unsupported reduction value: {}".format(reduction))


def mask_accuracy_with_logits(input, *args, **kwargs):
    """
    Measure accuracy with logit inputs.

    Pass through a sigmoid, binarize, then measure accuracy of predictions
    compared to ground truth target.

    See Also
    --------
    mask_accuracy
    """
    return mask_accuracy(torch.sigmoid(input), *args, **kwargs)


def mask_precision(input, target, threshold=0.5, ndim=None, reduction="mean"):
    """
    Measure precision of probability input.

    Binarize with a threshold, then measure precision compared to a ground truth target.

    Parameters
    ----------
    input : torch.Tensor
        Input tensor.
    target : torch.Tensor
        Target tensor, the same shape as ``input``.
    threshold : float, optional
        Threshold which entries in ``input`` and ``target`` must exceed to be
        binarised as the positive class. Default is ``0.5``.
    ndim : int or None
        Number of dimensions to keep. If ``None``, only the first (batch)
        dimension is kept and the rest are flattened. Default is ``None``.
    reduction : ``"none"`` or ``"mean"`` or ``"sum"``, optional
        Specifies the reduction to apply to the output:
        ``"none"`` | ``"mean"`` | ``"sum"``.
        ``"none"``: no reduction will be applied,
        ``"mean"``: the sum of the output will be divided by the number of
        elements in the output,
        ``"sum"``: the output will be summed.
        Default: ``"mean"``.

    Returns
    -------
    torch.Tensor
        The precision of ``input`` as compared to ``target`` after thresholding.
        The fraction of predicted positive cases, `input > 0.5`, which are
        true positive cases (`input > 0.5 and `target > 0.5`).
        If there are no predicted positives, the output is ``0`` if there are
        any positives to predict and ``1`` if there are none.
    """
    # Binarise and reshape masks
    input = _binarise_and_reshape(input, threshold=threshold, ndim=ndim)
    target = _binarise_and_reshape(target, threshold=threshold, ndim=ndim)

    # Measure true positives and total predicted positives
    true_p = (input & target).sum(-1)
    predicted_p = input.sum(-1)
    output = true_p.float() / predicted_p.float()
    # Handle division by 0: If there were no positives predicted, check whether
    # there were any to find.
    output[predicted_p == 0] = (true_p[predicted_p == 0] == 0).float()

    # Apply reduction
    if reduction == "none":
        return output
    elif reduction == "mean":
        return output.mean()
    elif reduction == "sum":
        return output.sum()
    else:
        raise ValueError("Unsupported reduction value: {}".format(reduction))


def mask_precision_with_logits(input, *args, **kwargs):
    """
    Measure precision of logit input.

    Pass through sigmoid, threshold, then measure precision.

    See Also
    --------
    mask_precision
    """
    return mask_precision(torch.sigmoid(input), *args, **kwargs)


def mask_recall(input, target, threshold=0.5, ndim=None, reduction="mean"):
    """
    Measure recall of probability input.

    Binarize with a threshold, then measure the recall compared to a ground truth target.

    Parameters
    ----------
    input : torch.Tensor
        Input tensor.
    target : torch.Tensor
        Target tensor, the same shape as ``input``.
    threshold : float, optional
        Threshold which entries in ``input`` and ``target`` must exceed to be
        binarised as the positive class. Default is ``0.5``.
    ndim : int or None
        Number of dimensions to keep. If ``None``, only the first (batch)
        dimension is kept and the rest are flattened. Default is ``None``.
    reduction : ``"none"`` or ``"mean"`` or ``"sum"``, optional
        Specifies the reduction to apply to the output:
        ``"none"`` | ``"mean"`` | ``"sum"``.
        ``"none"``: no reduction will be applied,
        ``"mean"``: the sum of the output will be divided by the number of
        elements in the output,
        ``"sum"``: the output will be summed.
        Default: ``"mean"``.

    Returns
    -------
    torch.Tensor
        The recall of ``input`` as compared to ``target`` after thresholding.
        The fraction of true positive cases, `target > 0.5`, which are
        true positive cases (`input > 0.5 and `target > 0.5`).
        If there are no true positives, the output is ``1``.
    """
    # Binarise and reshape masks
    input = _binarise_and_reshape(input, threshold=threshold, ndim=ndim)
    target = _binarise_and_reshape(target, threshold=threshold, ndim=ndim)

    # Measure true positives and actual positives
    true_p = (input & target).sum(-1)
    ground_truth_p = target.sum(-1)
    output = true_p.float() / ground_truth_p.float()
    # Handle division by 0: If there were no positives to find, all were found.
    output[ground_truth_p == 0] = 1

    # Apply reduction
    if reduction == "none":
        return output
    elif reduction == "mean":
        return output.mean()
    elif reduction == "sum":
        return output.sum()
    else:
        raise ValueError("Unsupported reduction value: {}".format(reduction))


def mask_recall_with_logits(input, *args, **kwargs):
    """
    Measure recall of logit input.

    Pass through sigmoid, binarize, then measure recall of ground truth target.

    See Also
    --------
    mask_recall
    """
    return mask_recall(torch.sigmoid(input), *args, **kwargs)


def mask_f1_score(input, target, reduction="mean", **kwargs):
    """
    Measure F1-score of probability input.

    Binarize, then measure the F1-score of the input vs ground truth target,

    Parameters
    ----------
    input : torch.Tensor
        Input tensor.
    target : torch.Tensor
        Target tensor, the same shape as ``input``.
    threshold : float, optional
        Threshold which entries in ``input`` and ``target`` must exceed to be
        binarised as the positive class. Default is ``0.5``.
    ndim : int or None
        Number of dimensions to keep. If ``None``, only the first (batch)
        dimension is kept and the rest are flattened. Default is ``None``.
    reduction : ``"none"`` or ``"mean"`` or ``"sum"``, optional
        Specifies the reduction to apply to the output:
        ``"none"`` | ``"mean"`` | ``"sum"``.
        ``"none"``: no reduction will be applied,
        ``"mean"``: the sum of the output will be divided by the number of
        elements in the output,
        ``"sum"``: the output will be summed.
        Default: ``"mean"``.

    Returns
    -------
    torch.Tensor
        The F1-score of ``input`` as compared to ``target`` after thresholding.
        The F1-score is the harmonic mean of precision and recall.

    See Also
    --------
    mask_precision
    mask_recall
    """
    precision = mask_precision(input, target, reduction="none", **kwargs)
    recall = mask_recall(input, target, reduction="none", **kwargs)
    sum_pr = precision + recall
    output = 2 * precision * recall / sum_pr
    # Handle division by 0: If both precision and recall are 0, f1 is 0 too.
    output[sum_pr == 0] = 0

    # Apply reduction
    if reduction == "none":
        return output
    elif reduction == "mean":
        return output.mean()
    elif reduction == "sum":
        return output.sum()
    else:
        raise ValueError("Unsupported reduction value: {}".format(reduction))


def mask_f1_score_with_logits(input, *args, **kwargs):
    """
    Measure F1-score of logit input.

    Convert logits to probabilities with sigmoid, apply a threshold, then
    measure the F1-score of the tensor as compared to ground truth.

    See Also
    --------
    mask_f1_score
    """
    return mask_f1_score(torch.sigmoid(input), *args, **kwargs)


def mask_jaccard_index(input, target, threshold=0.5, ndim=None, reduction="mean"):
    """
    Measure Jaccard Index from probabilities.

    Measure the Jaccard Index (intersection over union) of the input as
    compared to a ground truth target, after binarising with a threshold.

    Parameters
    ----------
    input : torch.Tensor
        Input tensor.
    target : torch.Tensor
        Target tensor, the same shape as ``input``.
    threshold : float, optional
        Threshold which entries in ``input`` and ``target`` must exceed to be
        binarised as the positive class. Default is ``0.5``.
    ndim : int or None
        Number of dimensions to keep. If ``None``, only the first (batch)
        dimension is kept and the rest are flattened. Default is ``None``.
    reduction : ``"none"`` or ``"mean"`` or ``"sum"``, optional
        Specifies the reduction to apply to the output:
        ``"none"`` | ``"mean"`` | ``"sum"``.
        ``"none"``: no reduction will be applied,
        ``"mean"``: the sum of the output will be divided by the number of
        elements in the output,
        ``"sum"``: the output will be summed.
        Default: ``"mean"``.

    Returns
    -------
    torch.Tensor
        The Jaccard Index of ``input`` as compared to ``target``.
        The Jaccard Index is the number of elements where both ``input`` and
        ``target`` exceed ``threshold``, divided by the number of elements where
        at least one of ``input`` and ``target`` exceeds ``threshold``.
    """
    # Binarise and reshape masks
    input = _binarise_and_reshape(input, threshold=threshold, ndim=ndim)
    target = _binarise_and_reshape(target, threshold=threshold, ndim=ndim)

    # Use bitwise operators to determine intersection and union of masks
    intersect = input & target
    union = input | target
    # Use number of pixels at which intersect and union are activated
    intersect = intersect.sum(-1)
    union = union.sum(-1)
    output = intersect.float() / union.float()
    # Handle division by 0: If there is no union, the two masks match completely.
    output[union == 0] = 1

    # Apply reduction
    if reduction == "none":
        return output
    elif reduction == "mean":
        return output.mean()
    elif reduction == "sum":
        return output.sum()
    else:
        raise ValueError("Unsupported reduction value: {}".format(reduction))


def mask_jaccard_index_with_logits(input, *args, **kwargs):
    """
    Measure Jaccard Index from logits.

    Convert logits to probabilities with sigmoid, apply a threshold, then
    measure the Jaccard Index (intersection over union) of the tensor as
    compared to ground truth.

    See Also
    --------
    mask_jaccard_index
    """
    return mask_jaccard_index(torch.sigmoid(input), *args, **kwargs)



#!/usr/bin/env python

import os
import sys
from shutil import rmtree

from setuptools import Command, find_packages, setup
from setuptools.command.test import test as TestCommand


def read(fname):
    with open(os.path.join(os.path.dirname(__file__), fname)) as f:
        return f.read()


def process_requirements(requirements):
    """Parse requirements, handling git repositories."""
    EGG_MARK = "#egg="
    required = []
    dependency_links = []
    for i_req, req in enumerate(requirements):
        if (
            req.startswith("-e git:")
            or req.startswith("-e git+")
            or req.startswith("git:")
            or req.startswith("git+")
        ):
            if EGG_MARK not in req:
                raise ValueError(
                    "Dependency to a git repository should have the format:\n"
                    "git+ssh://git@github.com/org/repo@ref#egg=package\n"
                    "But line {} contained:\n{}".format(i_req, req)
                )
            package_name = req[req.find(EGG_MARK) + len(EGG_MARK) :]
            required.append(package_name)
            dependency_links.append(req)
        else:
            required.append(req)
    return required, dependency_links


def process_requirements_file(fname):
    """Parse requirements from a given filename."""
    return process_requirements(read(fname).splitlines())


install_requires, dependency_links = process_requirements_file("requirements.txt")

extras_require = {}
for tag in ("dev", "docs", "test", "train", "plots"):
    extras_require[tag], extra_dep_links = process_requirements_file(
        "requirements-{}.txt".format(tag)
    )
    dependency_links += extra_dep_links


# If there are any extras, add a catch-all case that includes everything.
# This assumes that entries in extras_require are lists (not single strings).
if extras_require:
    extras_require["all"] = sorted({x for v in extras_require.values() for x in v})


# Import meta data from __meta__.py
#
# We use exec for this because __meta__.py runs its __init__.py first,
# __init__.py may assume the requirements are already present, but this code
# is being run during the `python setup.py install` step, before requirements
# are installed.
# https://packaging.python.org/guides/single-sourcing-package-version/
meta = {}
exec(read("echofilter/__meta__.py"), meta)


# Import the README and use it as the long-description.
# If your readme path is different, add it here.
possible_readme_names = ["README.rst", "README.md", "README.txt", "README"]

# Handle turning a README file into long_description
long_description = meta["description"]
readme_fname = ""
for fname in possible_readme_names:
    try:
        long_description = read(fname)
    except IOError:
        # doesn't exist
        continue
    else:
        # exists
        readme_fname = fname
        break

# Infer the content type of the README file from its extension.
# If the contents of your README do not match its extension, manually assign
# long_description_content_type to the appropriate value.
readme_ext = os.path.splitext(readme_fname)[1]
if readme_ext.lower() == ".rst":
    long_description_content_type = "text/x-rst"
elif readme_ext.lower() == ".md":
    long_description_content_type = "text/markdown"
else:
    long_description_content_type = "text/plain"


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest

        pytest.main(self.test_args)


setup(
    # Essential details on the package and its dependencies
    name=meta["name"],
    python_requires=">=3.6",
    version=meta["version"],
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
    package_dir={meta["name"]: os.path.join(".", meta["path"])},
    # If any package contains *.txt or *.rst files, include them:
    package_data={"echofilter": ["checkpoints.yaml"]},
    install_requires=install_requires,
    extras_require=extras_require,
    dependency_links=dependency_links,
    # Metadata to display on PyPI
    author=meta["author"],
    author_email=meta["author_email"],
    description=meta["description"],
    long_description=long_description,
    long_description_content_type=long_description_content_type,
    license=meta["license"],
    url=meta["url"],
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
    ],
    # Could also include keywords, download_url, project_urls, etc.
    entry_points={
        "console_scripts": [
            "echofilter=echofilter.__main__:main",
            "echofilter-train=echofilter.ui.train_cli:main",
            "echofilter-generate-shards=echofilter.generate_shards:main",
            "ev2csv=echofilter.ev2csv:main",
        ],
    },
    # Custom commands
    cmdclass={"test": PyTest},
)
